import os
import napari
from napari.utils.notifications import show_info

from qtpy.QtWidgets import QWidget, QPushButton, QLabel, QGridLayout, QSpinBox, QCheckBox
from nellie.feature_extraction.hierarchical import Hierarchy
from nellie.segmentation.labelling import Label
from nellie.segmentation.mocap_marking import Markers
from nellie.segmentation.networking import Network
from nellie.tracking.hu_tracking import HuMomentTracking
from nellie.tracking.voxel_reassignment import VoxelReassigner
from nellie.segmentation.filtering import Filter


class BatchMode(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer
        self.nellie.valid_files = None
        self.num_t = None

        self.im_info = None

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        self.all_preprocessed = False
        self.all_segmented = False
        self.all_marked = False
        self.all_tracked = False
        self.all_reassigned = False

        # Label above the spinner box
        self.channel_label = QLabel("Channel to analyze:")

        self.channel_input = QSpinBox()
        self.channel_input.setRange(0, 0)
        self.channel_input.setValue(0)
        self.channel_input.setEnabled(False)
        self.channel_input.valueChanged.connect(self.change_channel)

        # Label above the spinner box
        self.time_label = QLabel("Number of temporal frames:")

        self.time_input = QSpinBox()
        self.time_input.setRange(1, 1)
        self.time_input.setValue(1)
        self.time_input.setEnabled(False)
        self.time_input.valueChanged.connect(self.change_t)
        self.num_t = 1

        # Checkbox for 'Remove edges'
        self.remove_edges_checkbox = QCheckBox("Remove image edges")
        self.remove_edges_checkbox.setEnabled(False)
        self.remove_edges_checkbox.setToolTip(
            "Originally for Snouty deskewed images. If you see weird image edge artifacts, enable this.")

        self.run_all_button = QPushButton(text="Run all")
        self.run_all_button.clicked.connect(self.run_all)
        self.run_all_button.setEnabled(True)

        self.preprocess_button = QPushButton(text="Run preprocessing")
        self.preprocess_button.clicked.connect(self.run_preprocessing)
        self.preprocess_button.setEnabled(False)

        self.segment_button = QPushButton(text="Run segmentation")
        self.segment_button.clicked.connect(self.run_segmentation)
        self.segment_button.setEnabled(False)

        self.mocap_button = QPushButton(text="Run mocap marking")
        self.mocap_button.clicked.connect(self.run_mocap)
        self.mocap_button.setEnabled(False)

        self.track_button = QPushButton(text="Run tracking")
        self.track_button.clicked.connect(self.run_tracking)
        self.track_button.setEnabled(False)

        self.reassign_button = QPushButton(text="Run voxel reassignment")
        self.reassign_button.clicked.connect(self.run_reassignment)
        self.reassign_button.setEnabled(False)

        self.feature_export_button = QPushButton(text="Run feature export")
        self.feature_export_button.clicked.connect(self.run_feature_export)
        self.feature_export_button.setEnabled(False)

        # Add buttons
        self.layout.addWidget(self.channel_label, 0, 0)
        self.layout.addWidget(self.channel_input, 0, 1)
        self.layout.addWidget(self.time_label, 1, 0)
        self.layout.addWidget(self.time_input, 1, 1)
        self.layout.addWidget(self.remove_edges_checkbox, 2, 1)

        self.layout.addWidget(QLabel("Run full pipeline"), 42, 0, 1, 2)
        self.layout.addWidget(self.run_all_button, 43, 0, 1, 2)

        self.layout.addWidget(QLabel("Run individual steps / Visualize"), 44, 0, 1, 2)
        self.layout.addWidget(self.preprocess_button, 45, 0, 1, 2)
        self.layout.addWidget(self.segment_button, 46, 0, 1, 2)
        self.layout.addWidget(self.mocap_button, 47, 0, 1, 2)
        self.layout.addWidget(self.track_button, 48, 0, 1, 2)
        self.layout.addWidget(self.feature_export_button, 49, 0, 1, 2)
        self.layout.addWidget(self.reassign_button, 50, 0, 1, 2)

    def change_channel(self):
        if self.nellie.file_select.single:
            self.nellie.file_select.initialize_single_file()
        else:
            self.nellie.file_select.initialize_folder(self.nellie.file_select.filepath)

    def change_t(self):
        self.num_t = self.time_input.value()

    def run_all(self):
        self.run_preprocessing()
        self.run_segmentation()
        self.run_mocap()
        self.run_tracking()
        self.run_reassignment()
        self.run_feature_export()

    def run_feature_export(self):
        for file in self.nellie.valid_files:
            feature_export = Hierarchy(file, self.num_t)
            feature_export.run()
        self.enable_buttons()

    def run_preprocessing(self):
        for file in self.nellie.valid_files:
            try:
                preprocessing = Filter(file, self.num_t, remove_edges=self.remove_edges_checkbox.isChecked())
                preprocessing.run()
            except Exception as e:
                show_info(f"Error in preprocessing {file.filename}: {e}")
        self.enable_buttons()

    def run_segmentation(self):
        for file in self.nellie.valid_files:
            try:
                segmentation = Label(file, self.num_t)
                segmentation.run()
                networking = Network(file, self.num_t)
                networking.run()
            except Exception as e:
                show_info(f"Error in segmentation {file.filename}: {e}")
        self.enable_buttons()

    def run_mocap(self):
        for file in self.nellie.valid_files:
            try:
                mocap_marking = Markers(file, self.num_t)
                mocap_marking.run()
            except Exception as e:
                show_info(f"Error in mocap marking {file.filename}: {e}")
        self.enable_buttons()

    def run_tracking(self):
        for file in self.nellie.valid_files:
            try:
                hu_tracking = HuMomentTracking(file, self.num_t)
                hu_tracking.run()
            except Exception as e:
                show_info(f"Error in tracking {file.filename}: {e}")
        self.enable_buttons()

    def run_reassignment(self):
        for file in self.nellie.valid_files:
            try:
                vox_reassign = VoxelReassigner(file, self.num_t)
                vox_reassign.run()
            except Exception as e:
                show_info(f"Error in voxel reassignment {file.filename}: {e}")
        self.enable_buttons()

    def check_sections(self):
        if self.check_all_exists('im_frangi'):
            self.all_preprocessed = True
        if self.check_all_exists('im_instance_label'):
            self.all_segmented = True
        if self.check_all_exists('im_marker'):
            self.all_marked = True
        if self.check_all_exists('flow_vector_array'):
            self.all_tracked = True
        if self.check_all_exists('im_branch_label_reassigned'):
            self.all_reassigned = True

    def check_all_exists(self, filepath_name):
        for file in self.nellie.valid_files:
            filepath = file.pipeline_paths[filepath_name]
            if not os.path.exists(filepath):
                return False
        return True

    def enable_buttons(self):
        self.check_sections()
        self.preprocess_button.setEnabled(True)

        if self.all_preprocessed:
            self.segment_button.setEnabled(True)
        else:
            self.segment_button.setEnabled(False)

        if self.all_segmented:
            self.mocap_button.setEnabled(True)
        else:
            self.mocap_button.setEnabled(False)

        if self.all_marked:
            self.track_button.setEnabled(True)
        else:
            self.track_button.setEnabled(False)

        if self.all_tracked:
            self.reassign_button.setEnabled(True)
            self.feature_export_button.setEnabled(True)
        else:
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import os

from napari.utils.notifications import show_info
from qtpy.QtWidgets import QGridLayout, QWidget, QPushButton, QLabel, QSpinBox, QCheckBox

from nellie import logger
from nellie.tracking.all_tracks_for_label import LabelTracks
from nellie.utils.general import get_reshaped_image


class NellieVisualizer(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.scale = (1, 1, 1)

        self.im_memmap = None
        self.num_t = None

        self.im_branch_label_reassigned_layer = None
        self.im_branch_label_reassigned = None
        self.im_obj_label_reassigned_layer = None
        self.im_obj_label_reassigned = None
        self.im_marker_layer = None
        self.im_marker = None
        self.im_skel_relabelled_layer = None
        self.im_instance_label_layer = None
        self.frangi_layer = None
        self.im_skel_relabelled = None
        self.im_instance_label = None
        self.im_frangi = None

        # Raw im button
        self.raw_button = QPushButton(text="Open raw image")
        self.raw_button.clicked.connect(self.open_raw)
        self.raw_button.setEnabled(False)

        # Preprocess open button
        self.open_preprocess_button = QPushButton(text="Open preprocessed image")
        self.open_preprocess_button.clicked.connect(self.open_preprocess_image)
        self.open_preprocess_button.setEnabled(False)

        # Segment open button
        self.open_segment_button = QPushButton(text="Open segmentation images")
        self.open_segment_button.clicked.connect(self.open_segment_image)
        self.open_segment_button.setEnabled(False)

        # open im button
        self.open_mocap_button = QPushButton(text="Open mocap marker image")
        self.open_mocap_button.clicked.connect(self.open_mocap_image)
        self.open_mocap_button.setEnabled(False)

        # open reassign button
        self.open_reassign_button = QPushButton(text="Open reassigned labels image")
        self.open_reassign_button.clicked.connect(self.open_reassign_image)
        self.open_reassign_button.setEnabled(False)

        # Label above the spinner box
        self.skip_vox_label = QLabel("Track every N voxel. N=")

        self.skip_vox = QSpinBox()
        self.skip_vox.setRange(1, 10000)
        self.skip_vox.setValue(5)
        self.skip_vox.setEnabled(False)

        # Track all frames
        self.track_all_frames = QCheckBox("Track all frames' voxels")
        self.track_all_frames.setChecked(True)
        self.track_all_frames.setEnabled(False)

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        self.layout.addWidget(self.raw_button, 1, 0, 1, 1)
        self.layout.addWidget(self.open_preprocess_button, 2, 0, 1, 1)
        self.layout.addWidget(self.open_segment_button, 3, 0, 1, 1)
        self.layout.addWidget(self.open_mocap_button, 4, 0, 1, 1)
        self.layout.addWidget(self.open_reassign_button, 5, 0, 1, 1)
        self.layout.addWidget(self.skip_vox_label, 6, 0, 1, 1)
        self.layout.addWidget(self.skip_vox, 6, 1, 1, 1)
        self.layout.addWidget(self.track_all_frames, 7, 0, 1, 1)

    def post_init(self):
        if not self.check_for_raw():
            return
        self.num_t = self.nellie.processor.time_input.value()
        self.set_scale()
        self.viewer.scale_bar.visible = True
        self.viewer.scale_bar.unit = 'um'

    def check_3d(self):
        if not self.nellie.im_info.no_z and self.viewer.dims.ndim != 3:
            # ndimensions should be 3 for viewer
            self.viewer.dims.ndim = 3
            self.viewer.dims.ndisplay = 3

    def check_for_raw(self):
        self.raw_button.setEnabled(False)
        try:
            # todo implement lazy loading if wanted
            im_memmap = self.nellie.im_info.get_im_memmap(self.nellie.im_info.im_path)
            self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.nellie.im_info)
            self.raw_button.setEnabled(True)
            return True
        except Exception as e:
            logger.error(e)
            show_info(f"Could not open raw image: {e}")
            self.im_memmap = None
            return False

    def set_scale(self):
        dim_sizes = self.nellie.im_info.dim_sizes
        if self.nellie.im_info.no_z:
            self.scale = (dim_sizes['Y'], dim_sizes['X'])
        else:
            self.scale = (dim_sizes['Z'], dim_sizes['Y'], dim_sizes['X'])

    def open_preprocess_image(self):
        im_frangi = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_frangi'])
        self.im_frangi = get_reshaped_image(im_frangi, self.num_t, self.nellie.im_info)

        self.check_3d()

        self.frangi_layer = self.viewer.add_image(self.im_frangi, name='pre processed', colormap='turbo', scale=self.scale)
        self.frangi_layer.interpolation = 'nearest'

    def open_segment_image(self):
        im_instance_label = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_instance_label'])
        self.im_instance_label = get_reshaped_image(im_instance_label, self.num_t, self.nellie.im_info)
        im_skel_relabelled = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_skel_relabelled'])
        self.im_skel_relabelled = get_reshaped_image(im_skel_relabelled, self.num_t, self.nellie.im_info)

        self.check_3d()

        self.im_instance_label_layer = self.viewer.add_labels(self.im_instance_label, name='segmentation: objects',
                                                              opacity=1, scale=self.scale)
        self.im_skel_relabelled_layer = self.viewer.add_labels(self.im_skel_relabelled, name='segmentation: branches',
                                                               opacity=1, scale=self.scale)
        self.im_instance_label_layer.mouse_drag_callbacks.append(self.on_label_click)
        self.im_skel_relabelled_layer.mouse_drag_callbacks.append(self.on_label_click)

    def on_label_click(self, layer, event):
        # if flow_vector_array path does not point to an existing file, return
        if not os.path.exists(self.nellie.im_info.pipeline_paths['flow_vector_array']):
            return

        pos = None
        if event.button == 1 and event.is_dragging is False and 'Alt' in event.modifiers:
            scaled_pos = self.viewer.cursor.position
            pos = [scaled_pos[i+1] / self.scale[i] for i in range(len(scaled_pos)-1)]
            pos = (scaled_pos[0], *pos)
        if pos is None:
            return

        try:
            pos = tuple(int(pos_dim) for pos_dim in pos)
            label = layer.data[pos]
        except Exception as e:
            logger.error(e)
            return

        if label == 0:
            return

        if layer == self.im_instance_label_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        elif layer == self.im_skel_relabelled_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        elif layer == self.im_branch_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_branch_label_reassigned']
        elif layer == self.im_obj_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_obj_label_reassigned']
        else:
            return

        label_tracks = LabelTracks(im_info=self.nellie.im_info, num_t=self.num_t, label_im_path=label_path)
        label_tracks.initialize()
        all_tracks = []
        all_props = {}
        max_track_num = 0
        if self.track_all_frames.isChecked() and (
                layer == self.im_branch_label_reassigned_layer or layer == self.im_obj_label_reassigned_layer):
            for frame in range(self.num_t):
                tracks, track_properties = label_tracks.run(label_num=label, start_frame=frame, end_frame=None,
                                                            min_track_num=max_track_num,
                                                            skip_coords=self.skip_vox.value())
                all_tracks += tracks
                for property in track_properties.keys():
                    if property not in all_props.keys():
                        all_props[property] = []
                    all_props[property] += track_properties[property]
                if len(tracks) == 0:
                    break
                max_track_num = max([track[0] for track in tracks])+1
        else:
            all_tracks, all_props = label_tracks.run(label_num=label, start_frame=pos[0], end_frame=None,
                                                     skip_coords=self.skip_vox.value())
        if len(all_tracks) == 0:
            return
        self.viewer.add_tracks(all_tracks, properties=all_props, name=f'tracks: {label}', scale=self.scale)
        self.viewer.layers.selection.active = layer
        self.check_file_existence()

    def open_mocap_image(self):
        self.check_3d()

        im_marker = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_marker'])
        self.im_marker = get_reshaped_image(im_marker, self.num_t, self.nellie.im_info)
        self.im_marker_layer = self.viewer.add_image(self.im_marker, name='mocap markers', colormap='red',
                                                     blending='additive', contrast_limits=[0, 1], scale=self.scale)
        self.im_marker_layer.interpolation = 'nearest'

    def open_reassign_image(self):
        self.check_3d()

        im_branch_label_reassigned = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_branch_label_reassigned'])
        self.im_branch_label_reassigned = get_reshaped_image(im_branch_label_reassigned, self.num_t, self.nellie.im_info)
        self.im_branch_label_reassigned_layer = self.viewer.add_labels(self.im_branch_label_reassigned, name='reassigned branch voxels', scale=self.scale)
        self.im_branch_label_reassigned_layer.mouse_drag_callbacks.append(self.on_label_click)

        im_obj_label_reassigned = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_obj_label_reassigned'])
        self.im_obj_label_reassigned = get_reshaped_image(im_obj_label_reassigned, self.num_t, self.nellie.im_info)
        self.im_obj_label_reassigned_layer = self.viewer.add_labels(self.im_obj_label_reassigned, name='reassigned object voxels', scale=self.scale)
        self.im_obj_label_reassigned_layer.mouse_drag_callbacks.append(self.on_label_click)

    def open_raw(self):
        self.check_3d()
        self.raw_layer = self.viewer.add_image(self.im_memmap, name='raw', colormap='gray',
                                               blending='additive', scale=self.scale)
        # make 3d interpolation to nearest
        self.raw_layer.interpolation = 'nearest'

    def check_file_existence(self):
        # set all other buttons to disabled first
        self.open_preprocess_button.setEnabled(False)
        self.open_segment_button.setEnabled(False)
        self.open_mocap_button.setEnabled(False)
        self.open_reassign_button.setEnabled(False)

        frangi_path = self.nellie.im_info.pipeline_paths['im_frangi']
        if os.path.exists(frangi_path):
            self.open_preprocess_button.setEnabled(True)
        else:
            self.open_preprocess_button.setEnabled(False)
            return

        im_instance_label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        im_skel_relabelled_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        if os.path.exists(im_instance_label_path) and os.path.exists(im_skel_relabelled_path):
            self.open_segment_button.setEnabled(True)
            self.skip_vox.setEnabled(True)
            self.track_all_frames.setEnabled(True)
            self.viewer.help = 'Alt + click a label to see its tracks'
        else:
            self.open_segment_button.setEnabled(False)
            self.skip_vox.setEnabled(False)
            self.track_all_frames.setEnabled(False)
            return

        im_marker_path = self.nellie.im_info.pipeline_paths['im_marker']
        if os.path.exists(im_marker_path):
            self.open_mocap_button.setEnabled(True)
        else:
            self.open_mocap_button.setEnabled(False)
            return

        im_branch_label_path = self.nellie.im_info.pipeline_paths['im_branch_label_reassigned']
        if os.path.exists(im_branch_label_path):
            self.open_reassign_button.setEnabled(True)
        else:
            self.open_reassign_button.setEnabled(False)
            return


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



from nellie import xp, ndi, logger, device_type
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image
from nellie.utils.gpu_functions import otsu_threshold, triangle_threshold


class Label:
    def __init__(self, im_info: ImInfo,
                 num_t=None,
                 threshold: float = 0,
                 snr_cleaning=False):
        self.im_info = im_info
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]
        self.threshold = threshold
        self.snr_cleaning = snr_cleaning

        self.im_memmap = None
        self.frangi_memmap = None

        self.max_label_num = 0

        if not self.im_info.no_z:
            self.min_z_radius_um = min(self.im_info.dim_sizes['Z'], 0.2)

        self.semantic_mask_memmap = None
        self.instance_label_memmap = None
        self.shape = ()

        self.debug = {}

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for semantic segmentation.')
        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)

        frangi_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_frangi'])
        self.frangi_memmap = get_reshaped_image(frangi_memmap, self.num_t, self.im_info)
        self.shape = self.frangi_memmap.shape

        im_instance_label_path = self.im_info.pipeline_paths['im_instance_label']
        self.instance_label_memmap = self.im_info.allocate_memory(im_instance_label_path, shape=self.shape,
                                                                  dtype='int32',
                                                                  description='instance segmentation',
                                                                  return_memmap=True)

    def _get_labels(self, frame):
        ndim = 2 if self.im_info.no_z else 3
        footprint = ndi.generate_binary_structure(ndim, 1)

        triangle = 10 ** triangle_threshold(xp.log10(frame[frame > 0]))
        otsu, _ = otsu_threshold(xp.log10(frame[frame > 0]))
        otsu = 10 ** otsu
        min_thresh = min([triangle, otsu])

        mask = frame > min_thresh

        if not self.im_info.no_z:
            mask = ndi.binary_fill_holes(mask)

        if not self.im_info.no_z and self.im_info.dim_sizes['Z'] < self.min_z_radius_um:
            mask = ndi.binary_opening(mask, structure=xp.ones((2, 2, 2)))
        elif self.im_info.no_z:
            mask = ndi.binary_opening(mask, structure=xp.ones((2, 2)))

        labels, _ = ndi.label(mask, structure=footprint)
        # remove anything 4 pixels or under using bincounts
        areas = xp.bincount(labels.ravel())[1:]
        mask = xp.where(xp.isin(labels, xp.where(areas > 4)[0]+1), labels, 0) > 0
        labels, _ = ndi.label(mask, structure=footprint)
        return mask, labels

    def _get_subtraction_mask(self, original_frame, labels_frame):
        subtraction_mask = original_frame.copy()
        subtraction_mask[labels_frame > 0] = 0
        return subtraction_mask

    def _get_object_snrs(self, original_frame, labels_frame):
        logger.debug('Calculating object SNRs.')
        subtraction_mask = self._get_subtraction_mask(original_frame, labels_frame)
        unique_labels = xp.unique(labels_frame)
        extend_bbox_by = 1
        keep_labels = []
        for label in unique_labels:
            if label == 0:
                continue
            coords = xp.nonzero(labels_frame == label)
            z_coords, r_coords, c_coords = coords

            zmin, zmax = xp.min(z_coords), xp.max(z_coords)
            rmin, rmax = xp.min(r_coords), xp.max(r_coords)
            cmin, cmax = xp.min(c_coords), xp.max(c_coords)

            zmin, zmax = xp.clip(zmin - extend_bbox_by, 0, labels_frame.shape[0]), xp.clip(zmax + extend_bbox_by, 0,
                                                                                           labels_frame.shape[0])
            rmin, rmax = xp.clip(rmin - extend_bbox_by, 0, labels_frame.shape[1]), xp.clip(rmax + extend_bbox_by, 0,
                                                                                           labels_frame.shape[1])
            cmin, cmax = xp.clip(cmin - extend_bbox_by, 0, labels_frame.shape[2]), xp.clip(cmax + extend_bbox_by, 0,
                                                                                           labels_frame.shape[2])

            # only keep objects over 1 std from its surroundings
            local_intensity = subtraction_mask[zmin:zmax, rmin:rmax, cmin:cmax]
            local_intensity_mean = local_intensity[local_intensity > 0].mean()
            local_intensity_std = local_intensity[local_intensity > 0].std()
            label_intensity_mean = original_frame[coords].mean()
            intensity_cutoff = label_intensity_mean / (local_intensity_mean + local_intensity_std)
            if intensity_cutoff > 1:
                keep_labels.append(label)

        keep_labels = xp.asarray(keep_labels)
        labels_frame = xp.where(xp.isin(labels_frame, keep_labels), labels_frame, 0)
        return labels_frame

    def _run_frame(self, t):
        logger.info(f'Running semantic segmentation, volume {t}/{self.num_t - 1}')
        original_in_mem = xp.asarray(self.im_memmap[t, ...])
        frangi_in_mem = xp.asarray(self.frangi_memmap[t, ...])
        _, labels = self._get_labels(frangi_in_mem)
        if self.snr_cleaning:
            labels = self._get_object_snrs(original_in_mem, labels)
        labels[labels > 0] += self.max_label_num
        self.max_label_num = xp.max(labels)
        return labels

    def _run_segmentation(self):
        for t in range(self.num_t):
            labels = self._run_frame(t)
            if device_type == 'cuda':
                labels = labels.get()
            if self.im_info.no_t or self.num_t == 1:
                self.instance_label_memmap[:] = labels[:]
            else:
                self.instance_label_memmap[t, ...] = labels

            self.instance_label_memmap.flush()

    def run(self):
        logger.info('Running semantic segmentation.')
        self._get_t()
        self._allocate_memory()
        self._run_segmentation()


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_tests\deskewed-2023-07-13_14-58-28_000_wt_0_acquire.ome.tif"
    im_info = ImInfo(im_path)
    segment_unique = Label(im_info, num_t=2)
    segment_unique.run()



from nellie.feature_extraction.hierarchical import Hierarchy
from nellie.im_info.im_info import ImInfo
from nellie.segmentation.filtering import Filter
from nellie.segmentation.labelling import Label
from nellie.segmentation.mocap_marking import Markers
from nellie.segmentation.networking import Network
from nellie.tracking.hu_tracking import HuMomentTracking
from nellie.tracking.voxel_reassignment import VoxelReassigner


def run(im_path, num_t=None, remove_edges=True, ch=0):
    im_info = ImInfo(im_path, ch=ch)

    preprocessing = Filter(im_info, num_t, remove_edges=remove_edges)
    preprocessing.run()

    segmenting = Label(im_info, num_t)
    segmenting.run()

    networking = Network(im_info, num_t)
    networking.run()

    mocap_marking = Markers(im_info, num_t)
    mocap_marking.run()

    hu_tracking = HuMomentTracking(im_info, num_t)
    hu_tracking.run()

    vox_reassign = VoxelReassigner(im_info, num_t)
    vox_reassign.run()

    hierarchy = Hierarchy(im_info, num_t)
    hierarchy.run()

    return im_info


if __name__ == "__main__":
    # Single file run
    # im_path = r"/Users/austin/Downloads/test.tif"
    # im_path = r"/Users/austin/GitHub/nellie-simulations/motion/fission_fusion/outputs/fusion-std_fusion-std_512-t_0p5.ome.tif"
    # im_info = run(im_path, remove_edges=False, ch=0)

    # Directory bactch run
    import os
    top_dirs = [
        r"C:\Users\austin\GitHub\nellie-supplemental\comparisons\simulations\multi_grid\outputs",
        r"C:\Users\austin\GitHub\nellie-supplemental\comparisons\simulations\separation\outputs",
        r"C:\Users\austin\GitHub\nellie-supplemental\comparisons\simulations\px_sizes\outputs",
        ]
    ch = 0
    num_t = 1
    # get all non-folder files
    for top_dir in top_dirs:
        all_files = os.listdir(top_dir)
        all_files = [os.path.join(top_dir, file) for file in all_files if not os.path.isdir(os.path.join(top_dir, file))]
        all_files = [file for file in all_files if file.endswith('.tif')]
        for file_num, tif_file in enumerate(all_files):
            # for ch in range(1):
            print(f'Processing file {file_num + 1} of {len(all_files)}, channel {ch + 1} of 1')
            im_info = ImInfo(tif_file, ch=ch)
            if os.path.exists(im_info.pipeline_paths['im_skel_relabelled']):
                print(f'Already exists, skipping.')
                continue
            im_info = run(tif_file, remove_edges=False, ch=ch, num_t=num_t)



from nellie import logger, xp


def get_reshaped_image(im, num_t=None, im_info=None):
    logger.debug('Reshaping image.')
    im_to_return = im
    if im_info.no_z:
        ndim = 2
    else:
        ndim = 3
    if 'T' not in im_info.axes or (len(im_info.axes) > ndim and len(im_to_return.shape) == ndim):
        im_to_return = im_to_return[None, ...]
        logger.debug(f'Adding time dimension to image, shape is now {im_to_return.shape}.')
    elif num_t is not None:
        num_t = min(num_t, im_to_return.shape[0])
        im_to_return = im_to_return[:num_t, ...]
        logger.debug(f'{num_t} timepoints found, shape is now {im_to_return.shape}.')
    return im_to_return


def bbox(im):
    if len(im.shape) == 2:
        rows = xp.any(im, axis=1)
        cols = xp.any(im, axis=0)
        if (not rows.any()) or (not cols.any()):
            return 0, 0, 0, 0
        rmin, rmax = xp.where(rows)[0][[0, -1]]
        cmin, cmax = xp.where(cols)[0][[0, -1]]
        return int(rmin), int(rmax), int(cmin), int(cmax)

    elif len(im.shape) == 3:
        r = xp.any(im, axis=(1, 2))
        c = xp.any(im, axis=(0, 2))
        z = xp.any(im, axis=(0, 1))
        if (not r.any()) or (not c.any()) or (not z.any()):
            return 0, 0, 0, 0, 0, 0
        rmin, rmax = xp.where(r)[0][[0, -1]]
        cmin, cmax = xp.where(c)[0][[0, -1]]
        zmin, zmax = xp.where(z)[0][[0, -1]]
        return int(rmin), int(rmax), int(cmin), int(cmax), int(zmin), int(zmax)

    else:
        print("Image not 2D or 3D... Cannot get bounding box.")
        return None



import os
import shutil
import tempfile
from nellie import ImInfo
# todo tests for different filetypes or input types... Think of others

def test_iminfo_get_metadata_5d():
    im_path = 'data/tczyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert iminfo.axes == 'TCZYX'
    assert iminfo.shape == (7, 3, 5, 167, 439)
    shutil.rmtree(output_dir)


def test_iminfo_get_metadata_3d():
    im_path = 'data/tyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert iminfo.axes == 'TYX'
    assert iminfo.shape == (7, 167, 439)
    shutil.rmtree(output_dir)


def test_iminfo_get_dim_sizes_5d():
    im_path = 'data/tczyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert iminfo.dim_sizes['X'] is None
    assert iminfo.dim_sizes['Y'] is None
    assert iminfo.dim_sizes['Z'] is None
    assert iminfo.dim_sizes['T'] is None
    assert iminfo.dim_sizes['C'] == 1
    shutil.rmtree(output_dir)


def test_iminfo_get_dim_sizes_3d():
    im_path = 'data/tyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert iminfo.dim_sizes['X'] is None
    assert iminfo.dim_sizes['Y'] is None
    assert iminfo.dim_sizes['Z'] is None
    assert iminfo.dim_sizes['T'] is None
    assert iminfo.dim_sizes['C'] == 1
    shutil.rmtree(output_dir)


def test_iminfo_create_output_dirs_5d():
    im_path = 'data/tczyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert os.path.isdir(iminfo.output_dirpath)
    assert os.path.isdir(iminfo.output_images_dirpath)
    assert os.path.isdir(iminfo.output_pickles_dirpath)
    assert os.path.isdir(iminfo.output_csv_dirpath)
    shutil.rmtree(output_dir)


def test_iminfo_create_output_dirs_3d():
    im_path = 'data/tyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir)
    assert os.path.isdir(iminfo.output_dirpath)
    assert os.path.isdir(iminfo.output_images_dirpath)
    assert os.path.isdir(iminfo.output_pickles_dirpath)
    assert os.path.isdir(iminfo.output_csv_dirpath)
    shutil.rmtree(output_dir)


def test_iminfo_init_5d():
    im_path = 'data/tczyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir, ch=1, dim_sizes={'X': 0.1, 'Y': 0.2, 'Z': 0.3, 'T': 0.4, 'C': 1})

    # test init method
    assert iminfo.im_path == im_path
    assert iminfo.ch == 1
    assert iminfo.dim_sizes['X'] == 0.1
    assert iminfo.dim_sizes['Y'] == 0.2
    assert iminfo.dim_sizes['Z'] == 0.3
    assert iminfo.dim_sizes['T'] == 0.4
    assert iminfo.dim_sizes['C'] == 1
    shutil.rmtree(output_dir)


def test_iminfo_init_3d():
    im_path = 'data/tyx_ex.tif'
    output_dir = tempfile.TemporaryDirectory().name
    iminfo = ImInfo(im_path, output_dir, ch=1, dim_sizes={'X': 0.1, 'Y': 0.2, 'T': 0.4, 'C': 1})

    # test init method
    assert iminfo.im_path == im_path
    assert iminfo.ch == 1
    assert iminfo.dim_sizes['X'] == 0.1
    assert iminfo.dim_sizes['Y'] == 0.2
    assert iminfo.dim_sizes['T'] == 0.4
    assert iminfo.dim_sizes['C'] == 1
    shutil.rmtree(output_dir)



import numpy as np
from scipy.spatial import cKDTree

from nellie import logger
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image


class FlowInterpolator:
    def __init__(self, im_info: ImInfo, num_t=None, max_distance_um=0.5, forward=True):
        self.im_info = im_info
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]

        if self.im_info.no_z:
            self.scaling = (im_info.dim_sizes['Y'], im_info.dim_sizes['X'])
        else:
            self.scaling = (im_info.dim_sizes['Z'], im_info.dim_sizes['Y'], im_info.dim_sizes['X'])

        self.max_distance_um = max_distance_um * im_info.dim_sizes['T']
        self.max_distance_um = np.max([self.max_distance_um, 0.5])

        self.forward = forward

        self.shape = ()

        self.im_memmap = None
        self.flow_vector_array = None

        # caching
        self.current_t = None
        self.check_rows = None
        self.check_coords = None
        self.current_tree = None

        self.debug = None
        self._initialize()

    def _allocate_memory(self):
        logger.debug('Allocating memory for mocap marking.')

        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)
        self.shape = self.im_memmap.shape

        flow_vector_array_path = self.im_info.pipeline_paths['flow_vector_array']
        self.flow_vector_array = np.load(flow_vector_array_path)

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _get_nearby_coords(self, t, coords):
        # using a ckdtree, check for any nearby coords from coord
        if self.current_t != t:
            self.current_tree = cKDTree(self.check_coords * self.scaling)
        scaled_coords = np.array(coords) * self.scaling
        # get all coords and distances within the radius of the coord
        # good coords are non-nan
        good_coords = np.where(~np.isnan(scaled_coords[:, 0]))[0]
        nearby_idxs = self.current_tree.query_ball_point(scaled_coords[good_coords], self.max_distance_um, p=2)
        if len(nearby_idxs) == 0:
            return [], []
        k_all = [len(nearby_idxs[i]) for i in range(len(nearby_idxs))]
        max_k = np.max(k_all)
        if max_k == 0:
            return [], []
        distances, nearby_idxs = self.current_tree.query(scaled_coords[good_coords], k=max_k, p=2, workers=-1)
        # if the first index is scalar, wrap the whole list in another list
        if len(distances.shape) == 1:
            distances = [distances]
            nearby_idxs = [nearby_idxs]
        distance_return = [[] for _ in range(len(coords))]
        nearby_idxs_return = [[] for _ in range(len(coords))]
        pos = 0
        for i in range(len(distances)):
            if i not in good_coords:
                continue
            distance_return[i] = distances[pos][:k_all[pos]]
            nearby_idxs_return[i] = nearby_idxs[pos][:k_all[pos]]
            pos += 1
        return nearby_idxs_return, distance_return

    def _get_vector_weights(self, nearby_idxs, distances_all):
        weights_all = []
        for i in range(len(nearby_idxs)):
            # lowest cost should be most highly weighted
            cost_weights = -self.check_rows[nearby_idxs[i], -1]

            if len(distances_all[i]) == 0:
                weights_all.append(None)
                continue

            if np.min(distances_all[i]) == 0:
                distance_weights = (distances_all[i] == 0) * 1.0
            else:
                distance_weights = 1 / distances_all[i]

            weights = cost_weights * distance_weights
            weights -= np.min(weights) - 1
            weights /= np.sum(weights)
            weights_all.append(weights)
        return weights_all

    def _get_final_vector(self, nearby_idxs, weights_all):
        if self.im_info.no_z:
            final_vectors = np.zeros((len(nearby_idxs), 2))
        else:
            final_vectors = np.zeros((len(nearby_idxs), 3))
        for i in range(len(nearby_idxs)):
            if weights_all[i] is None:
                final_vectors[i] = np.nan
                continue
            if self.im_info.no_z:
                vectors = self.check_rows[nearby_idxs[i], 3:5]
            else:
                vectors = self.check_rows[nearby_idxs[i], 4:7]
            if len(weights_all[i].shape) == 0:
                final_vectors[i] = vectors[0]
            else:
                weighted_vectors = vectors * weights_all[i][:, None]
                final_vectors[i] = np.sum(weighted_vectors, axis=0)  # already normalized by weights
        return final_vectors

    def interpolate_coord(self, coords, t):
        # interpolate the flow vector at the coordinate at time t, either forward in time or backward in time.
        # For forward, simply find nearby LMPs, interpolate based on distance-weighted vectors
        # For backward, get coords from t-1 + vector, then find nearby coords from that, and interpolate based on distance-weighted vectors
        if self.current_t != t:
            if self.forward:
                # check_rows will be all rows where the self.flow_vector_array's 0th column is equal to t
                self.check_rows = self.flow_vector_array[np.where(self.flow_vector_array[:, 0] == t)[0], :]
                if self.im_info.no_z:
                    self.check_coords = self.check_rows[:, 1:3]
                else:
                    self.check_coords = self.check_rows[:, 1:4]
            else:
                # check_rows will be all rows where the self.flow_vector_array's 0th columns is equal to t-1
                self.check_rows = self.flow_vector_array[np.where(self.flow_vector_array[:, 0] == t - 1)[0], :]
                # check coords will be the coords + vector
                if self.im_info.no_z:
                    self.check_coords = self.check_rows[:, 1:3] + self.check_rows[:, 3:5]
                else:
                    self.check_coords = self.check_rows[:, 1:4] + self.check_rows[:, 4:7]

        nearby_idxs, distances_all = self._get_nearby_coords(t, coords)
        self.current_t = t

        if nearby_idxs is None:
            return None

        weights_all = self._get_vector_weights(nearby_idxs, distances_all)
        final_vectors = self._get_final_vector(nearby_idxs, weights_all)

        return final_vectors

    def _initialize(self):
        if self.im_info.no_t:
            return
        self._get_t()
        self._allocate_memory()


def interpolate_all_forward(coords, start_t, end_t, im_info, min_track_num=0, max_distance_um=0.5):
    flow_interpx = FlowInterpolator(im_info, forward=True, max_distance_um=max_distance_um)
    tracks = []
    track_properties = {'frame_num': []}
    frame_range = np.arange(start_t, end_t)
    for t in frame_range:
        final_vector = flow_interpx.interpolate_coord(coords, t)
        if final_vector is None or len(final_vector) == 0:
            continue
        for coord_num, coord in enumerate(coords):
            if np.all(np.isnan(final_vector[coord_num])):
                coords[coord_num] = np.nan
                continue
            if t == frame_range[0]:
                if im_info.no_z:
                    tracks.append([coord_num + min_track_num, frame_range[0], coord[0], coord[1]])
                else:
                    tracks.append([coord_num + min_track_num, frame_range[0], coord[0], coord[1], coord[2]])
                track_properties['frame_num'].append(frame_range[0])

            track_properties['frame_num'].append(t + 1)
            if im_info.no_z:
                coords[coord_num] = np.array([coord[0] + final_vector[coord_num][0],
                                              coord[1] + final_vector[coord_num][1]])
                tracks.append([coord_num + min_track_num, t + 1, coord[0], coord[1]])
            else:
                coords[coord_num] = np.array([coord[0] + final_vector[coord_num][0],
                                              coord[1] + final_vector[coord_num][1],
                                              coord[2] + final_vector[coord_num][2]])
                tracks.append([coord_num + min_track_num, t + 1, coord[0], coord[1], coord[2]])
    return tracks, track_properties


def interpolate_all_backward(coords, start_t, end_t, im_info, min_track_num=0, max_distance_um=0.5):
    flow_interpx = FlowInterpolator(im_info, forward=False, max_distance_um=max_distance_um)
    tracks = []
    track_properties = {'frame_num': []}
    frame_range = list(np.arange(end_t, start_t + 1))[::-1]
    for t in frame_range:
        final_vector = flow_interpx.interpolate_coord(coords, t)
        if final_vector is None or len(final_vector) == 0:
            continue
        for coord_num, coord in enumerate(coords):
            # if final_vector[coord_num] is all nan, skip
            if np.all(np.isnan(final_vector[coord_num])):
                coords[coord_num] = np.nan
                continue
            if t == frame_range[0]:
                if im_info.no_z:
                    tracks.append([coord_num + min_track_num, frame_range[0], coord[0], coord[1]])
                else:
                    tracks.append([coord_num + min_track_num, frame_range[0], coord[0], coord[1], coord[2]])
                track_properties['frame_num'].append(frame_range[0])
            if im_info.no_z:
                coords[coord_num] = np.array([coord[0] - final_vector[coord_num][0],
                                              coord[1] - final_vector[coord_num][1]])
                tracks.append([coord_num + min_track_num, t - 1, coord[0], coord[1]])
            else:
                coords[coord_num] = np.array([coord[0] - final_vector[coord_num][0],
                                              coord[1] - final_vector[coord_num][1],
                                              coord[2] - final_vector[coord_num][2]])
                tracks.append([coord_num + min_track_num, t - 1, coord[0], coord[1], coord[2]])
            track_properties['frame_num'].append(t - 1)
    return tracks, track_properties


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_smorgasbord\deskewed-iono_pre.ome.tif"
    im_info = ImInfo(im_path)
    label_memmap = im_info.get_im_memmap(im_info.pipeline_paths['im_instance_label'])
    label_memmap = get_reshaped_image(label_memmap, im_info=im_info)
    im_memmap = im_info.get_im_memmap(im_info.im_path)
    im_memmap = get_reshaped_image(im_memmap, im_info=im_info)

    import napari
    viewer = napari.Viewer()
    start_frame = 0
    # going backwards
    coords = np.argwhere(label_memmap[0] > 0).astype(float)
    # get 100 random coords
    # np.random.seed(0)
    # coords = coords[np.random.choice(coords.shape[0], 10000, replace=False), :].astype(float)
    # x in range 450-650
    # y in range 600-750
    new_coords = []
    for coord in coords:
        if 450 < coord[-1] < 650 and 600 < coord[-2] < 750:
            new_coords.append(coord)
    coords = np.array(new_coords[::1])
    tracks, track_properties = interpolate_all_forward(coords, start_frame, 3, im_info)

    viewer.add_image(im_memmap)
    viewer.add_tracks(tracks, properties=track_properties, name='tracks')



import os
import pickle

import numpy as np
from qtpy.QtWidgets import QGridLayout, QComboBox, QCheckBox, QPushButton, QLabel, QTableWidget, QTableWidgetItem, \
    QSpinBox, QDoubleSpinBox, QWidget
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from napari.utils.notifications import show_info
import pandas as pd
from nellie.utils.general import get_reshaped_image
import datetime


class NellieAnalysis(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.num_t = None
        self.canvas = FigureCanvasQTAgg()
        self.canvas.figure.set_layout_engine("constrained")

        self.scale = (1, 1, 1)

        self.log_scale = False
        self.is_median = False
        self.mean_median_toggle = None
        self.overlay_button = None
        self.match_t_toggle = None
        self.match_t = False
        self.hist_min = None
        self.hist_max = None
        self.num_bins = None
        self.hist_reset = True

        self.click_match_table = None

        self.voxel_df = None
        self.node_df = None
        self.branch_df = None
        self.organelle_df = None
        self.image_df = None

        self.all_attr_data = None
        self.attr_data = None
        self.time_col = None
        self.adjacency_maps = None
        self.data_to_plot = None
        self.mean = np.nan
        self.std = np.nan
        self.median = np.nan
        self.iqr = np.nan
        self.perc75 = np.nan
        self.perc25 = np.nan

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        self.dropdown = None
        self.dropdown_attr = None
        self.log_scale_checkbox = None
        self.selected_level = None

        self.label_mask = None
        self.label_mask_layer = None
        self.label_coords = []

        self.click_match_texts = []

        self.layout_anchors = {
            'dropdown': (0, 0),
            'canvas': (1, 0),
            'table': (50, 0)
        }

        self.df = None

    def post_init(self):
        self.num_t = self.nellie.processor.time_input.value()

        self._create_dropdown_selection()

        self.log_scale_checkbox = QCheckBox("Log scale")
        self.log_scale_checkbox.stateChanged.connect(self.on_log_scale)

        self.mean_median_toggle = QCheckBox("Median view")
        self.mean_median_toggle.stateChanged.connect(self.toggle_mean_med)

        self.overlay_button = QPushButton("Overlay mask")
        self.overlay_button.clicked.connect(self.overlay)

        self.match_t_toggle = QCheckBox("Timepoint data")
        self.match_t_toggle.stateChanged.connect(self.toggle_match_t)
        self.match_t_toggle.setEnabled(False)
        self.viewer.dims.events.current_step.connect(self.on_t_change)

        self.hist_min = QDoubleSpinBox()
        self.hist_min.setEnabled(False)
        self.hist_min.valueChanged.connect(self.on_hist_change)
        self.hist_min.setDecimals(4)

        self.hist_max = QDoubleSpinBox()
        self.hist_max.setEnabled(False)
        self.hist_max.valueChanged.connect(self.on_hist_change)
        self.hist_max.setDecimals(4)

        self.num_bins = QSpinBox()
        self.num_bins.setRange(1, 100)
        self.num_bins.setValue(10)
        self.num_bins.setEnabled(False)
        self.num_bins.valueChanged.connect(self.on_hist_change)

        self.export_data_button = QPushButton("Export graph data")
        self.export_data_button.clicked.connect(self.export_data)

        self.save_graph_button = QPushButton("Save graph")
        self.save_graph_button.clicked.connect(self.save_graph)

        canvas_width = 4
        canvas_height = 6
        self.layout.addWidget(self.canvas, self.layout_anchors['canvas'][0], self.layout_anchors['canvas'][1],
                              canvas_height, canvas_width)
        self.layout.addWidget(self.export_data_button, self.layout_anchors['canvas'][0]+canvas_height-2, 49, 1, 2)
        self.layout.addWidget(self.save_graph_button, self.layout_anchors['canvas'][0]+canvas_height-1, 49, 1, 2)
        self.layout.addWidget(QLabel("Max"), self.layout_anchors['canvas'][0], 49)
        self.layout.addWidget(self.hist_max, self.layout_anchors['canvas'][0], 50)
        self.layout.addWidget(QLabel("Min"), self.layout_anchors['canvas'][0]+1, 49)
        self.layout.addWidget(self.hist_min, self.layout_anchors['canvas'][0]+1, 50)
        self.layout.addWidget(QLabel("Bins"), self.layout_anchors['canvas'][0]+3, 49)
        self.layout.addWidget(self.num_bins, self.layout_anchors['canvas'][0]+3, 50)

        self.layout.addWidget(self.log_scale_checkbox, self.layout_anchors['table'][0]-1, 0)
        self.layout.addWidget(self.mean_median_toggle, self.layout_anchors['table'][0]-1, 1)
        self.layout.addWidget(self.overlay_button, self.layout_anchors['table'][0]-1, 2)
        self.layout.addWidget(self.match_t_toggle, self.layout_anchors['table'][0]-1, 3)

        if self.nellie.im_info.no_z:
            self.scale = (self.nellie.im_info.dim_sizes['Y'], self.nellie.im_info.dim_sizes['X'])
        else:
            self.scale = (self.nellie.im_info.dim_sizes['Z'], self.nellie.im_info.dim_sizes['Y'], self.nellie.im_info.dim_sizes['X'])
        self.viewer.scale_bar.visible = True
        self.viewer.scale_bar.unit = 'um'

        self.dropdown.setCurrentIndex(1)
        self.dropdown_attr.setCurrentIndex(1)


    def export_data(self):
        dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        export_dir = self.nellie.im_info.graph_dir
        if not os.path.exists(export_dir):
            os.makedirs(export_dir)
        text = f"{dt}-{self.selected_level}-{self.dropdown_attr.currentText()}"
        if self.match_t:
            current_t = self.viewer.dims.current_step[0]
            text += f"_T{current_t}"
            timepoints = self.time_col[self.df['t'] == current_t]
        else:
            timepoints = self.time_col
        text += self.nellie.im_info.basename_no_ext
        export_path = os.path.join(export_dir, f"{text}.csv")
        df_to_save = pd.DataFrame({'t': timepoints, self.dropdown_attr.currentText(): self.attr_data})
        df_to_save.to_csv(export_path)

        # append time column
        show_info(f"Data exported to {export_path}")

    def save_graph(self):
        dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        export_dir = self.nellie.im_info.graph_dir
        if not os.path.exists(export_dir):
            os.makedirs(export_dir)
        text = f"{dt}-{self.selected_level}_{self.dropdown_attr.currentText()}"
        if self.match_t:
            text += f"_T{self.viewer.dims.current_step[0]}"
        text += self.nellie.im_info.basename_no_ext
        export_path = os.path.join(export_dir, f"{text}.png")
        self.canvas.figure.savefig(export_path, dpi=300)
        show_info(f"Graph saved to {export_path}")

    def on_hist_change(self, event):
        self.plot_data(self.dropdown_attr.currentText())

    def get_index(self, layer, event):

        # get the coordinates of where the mouse is hovering
        pos = self.viewer.cursor.position
        matched_row = None
        if self.nellie.im_info.no_z:
            t, y, x = int(np.round(pos[0])), int(np.round(pos[1])), int(np.round(pos[2]))
        else:
            # todo, gonna be more complicated than this I think
            t, z, y, x = int(np.round(pos[0])), int(np.round(pos[1])), int(np.round(pos[2])), int(np.round(pos[3]))
        # find corresponding coord in self.label_coords
        if self.nellie.im_info.no_z:
            t_coords = self.label_coords[t]
            match = np.where((t_coords[:, 0] == y) & (t_coords[:, 1] == x))
            if len(match[0]) == 0:
                return
            else:
                matched_row = match[0][0]
            # show_info(f"Matched csv index: {matched_idx}, at T,Y,X: {t, y, x}")
        else:
            coord = self.label_coords[t][z]
        if matched_row is None:
            return
        # get the value of self.voxel_df_idxs[self.voxel_time_col == t] at the matched_idx row
        voxel_idx = self.voxel_df[self.voxel_df.columns[0]][self.voxel_df['t'] == t].iloc[matched_row]
        node_row = np.where(self.adjacency_maps['v_n'][t][matched_row])[0]
        node_idx = self.node_df[self.node_df.columns[0]][self.node_df['t'] == t].iloc[node_row].values
        branch_row = np.where(self.adjacency_maps['v_b'][t][matched_row])[0][0]
        branch_idx = self.branch_df[self.branch_df.columns[0]][self.branch_df['t'] == t].iloc[branch_row]
        organelle_row = np.where(self.adjacency_maps['v_o'][t][matched_row])[0][0]
        organelle_idx = self.organelle_df[self.organelle_df.columns[0]][self.organelle_df['t'] == t].iloc[organelle_row]
        image_row = np.where(self.adjacency_maps['v_i'][t][matched_row])[0][0]
        image_idx = self.image_df[self.image_df.columns[0]][self.image_df['t'] == t].iloc[image_row]

        self.click_match_table = QTableWidget()
        self.click_match_table.setRowCount(1)
        items = [f"{voxel_idx}", f"{node_idx}", f"{branch_idx}", f"{organelle_idx}", f"{image_idx}"]
        self.click_match_table.setColumnCount(len(items))
        self.click_match_table.setHorizontalHeaderLabels(["Voxel", "Nodes", "Branch", "Organelle", "Image"])
        for i, item in enumerate(items):
            self.click_match_table.setItem(0, i, QTableWidgetItem(item))
        self.layout.addWidget(self.click_match_table, self.layout_anchors['table'][0], self.layout_anchors['table'][1], 1, 4)
        self.click_match_table.setVerticalHeaderLabels([f"{t, y, x}\nCSV row"])


    def overlay(self):
        if self.label_mask is None:
            label_mask = self.nellie.im_info.get_im_memmap(self.nellie.im_info.pipeline_paths['im_instance_label'])
            self.label_mask = (get_reshaped_image(label_mask, self.num_t, self.nellie.im_info) > 0).astype(float)
            # self.label_mask_layer = self.viewer.add_image(self.label_mask, name='objects', opacity=1, colormap='turbo')
            if self.num_t is None:
                self.num_t = self.label_mask.shape[0]
                t_max = self.num_t
            else:
                t_max = min(self.label_mask.shape[0], self.num_t)
            for t in range(t_max):
                self.label_coords.append(np.argwhere(self.label_mask[t]))
                self.label_mask[t] *= np.nan

        if self.adjacency_maps is None:
            pkl_path = self.nellie.im_info.pipeline_paths['adjacency_maps']
            # load pkl file
            with open(pkl_path, 'rb') as f:
                self.adjacency_maps = pickle.load(f)
        if self.attr_data is None:
            return

        for t in range(self.num_t):
            t_attr_data = self.all_attr_data[self.time_col == t].astype(float)
            if len(t_attr_data) == 0:
                continue
            if self.selected_level == 'voxel':
                self.label_mask[t][tuple(self.label_coords[t].T)] = t_attr_data
                continue
            elif self.selected_level == 'node':
                adjacency_mask = np.array(self.adjacency_maps['n_v'][t])
            elif self.selected_level == 'branch':
                adjacency_mask = np.array(self.adjacency_maps['b_v'][t])
            elif self.selected_level == 'organelle':
                adjacency_mask = np.array(self.adjacency_maps['o_v'][t])
            elif self.selected_level == 'image':
                adjacency_mask = np.array(self.adjacency_maps['i_v'][t])
            else:
                return
            reshaped_t_attr = t_attr_data.values.reshape(-1, 1)
            # adjacency_mask = np.array(self.adjacency_maps['n_v'][t])
            attributed_voxels = adjacency_mask * reshaped_t_attr
            attributed_voxels[~adjacency_mask] = np.nan
            voxel_attributes = np.nanmean(attributed_voxels, axis=0)
            self.label_mask[t][tuple(self.label_coords[t].T)] = voxel_attributes

        layer_name = f'{self.selected_level} {self.dropdown_attr.currentText()}'
        if 'reassigned' in self.dropdown_attr.currentText():
            # make the label_mask_layer a label layer
            self.label_mask_layer = self.viewer.add_labels(self.label_mask.astype('uint64'), scale=self.scale, name=layer_name)
        else:
            self.label_mask_layer = self.viewer.add_image(self.label_mask, name=layer_name, opacity=1,
                                                          colormap='turbo', scale=self.scale)
            perc98 = np.nanpercentile(self.attr_data, 98)
            min_val = np.nanmin(self.attr_data) - (np.abs(np.nanmin(self.attr_data)) * 0.01)
            if min_val == perc98:
                perc98 = min_val + (np.abs(min_val) * 0.01)
            contrast_limits = [min_val, perc98]
            self.label_mask_layer.contrast_limits = contrast_limits
        self.label_mask_layer.name = layer_name
        if not self.nellie.im_info.no_z:
            # if the layer isn't in 3D view, make it 3d view
            self.viewer.dims.ndisplay = 3
            self.label_mask_layer.interpolation3d = 'nearest'
        self.label_mask_layer.refresh()
        self.label_mask_layer.mouse_drag_callbacks.append(self.get_index)
        self.match_t_toggle.setEnabled(True)
        self.viewer.reset_view()

    def on_t_change(self, event):
        if self.match_t:
            self.on_attr_selected(self.dropdown_attr.currentIndex())

    def toggle_match_t(self, state):
        if state == 2:
            self.match_t = True
        else:
            self.match_t = False
        self.on_attr_selected(self.dropdown_attr.currentIndex())

    def toggle_mean_med(self, state):
        if state == 2:
            self.is_median = True
        else:
            self.is_median = False
        self.on_attr_selected(self.dropdown_attr.currentIndex())

    def _create_dropdown_selection(self):
        # Create the dropdown menu
        self.dropdown = QComboBox()

        # Add options to the dropdown
        options = ['none', 'voxel', 'node', 'branch', 'organelle', 'image']
        for option in options:
            self.dropdown.addItem(option)

        # Add the dropdown to the layout
        self.layout.addWidget(self.dropdown, self.layout_anchors['dropdown'][0], self.layout_anchors['dropdown'][1])

        # Connect the dropdown's signal to a method to handle selection changes
        self.dropdown.currentIndexChanged.connect(self.on_level_selected)

        self.get_csvs()

    def get_csvs(self):
        self.voxel_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_voxels'])
        self.node_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_nodes'])
        self.branch_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_branches'])
        self.organelle_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_components'])
        self.image_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_image'])

        # self.voxel_time_col = voxel_df['t']
        # self.voxel_df_idxs = voxel_df[voxel_df.columns[0]]

    def on_level_selected(self, index):
        # This method is called whenever a radio button is selected
        # 'button' parameter is the clicked radio button
        self.selected_level = self.dropdown.itemText(index)
        if self.selected_level == 'voxel':
            self.df = self.voxel_df
        elif self.selected_level == 'node':
            self.df = self.node_df
        elif self.selected_level == 'branch':
            self.df = self.branch_df
        elif self.selected_level == 'organelle':
            self.df = self.organelle_df
        elif self.selected_level == 'image':
            self.df = self.image_df
        else:
            return
        # self.df = pd.read_csv(csv_path)
        self.dropdown_attr = QComboBox()
        # add a None option
        self.dropdown_attr.addItem("None")
        for col in self.df.columns[::-1]:
            self.dropdown_attr.addItem(col)
        self.layout.addWidget(self.dropdown_attr, self.layout_anchors['dropdown'][0], self.layout_anchors['dropdown'][1] + 1)
        self.dropdown_attr.currentIndexChanged.connect(self.on_attr_selected)

    def on_attr_selected(self, index):
        self.hist_reset = True

        selected_attr = self.dropdown_attr.itemText(index)
        if selected_attr == "None":
            # clear the canvas
            self.canvas.figure.clear()
            self.canvas.draw()
            return

        self.all_attr_data = self.df[selected_attr]
        self.time_col = self.df['t']
        if self.match_t:
            t = self.viewer.dims.current_step[0]
            self.attr_data = self.df[self.df['t'] == t][selected_attr]
        else:
            self.attr_data = self.df[selected_attr]
        self.get_stats()
        self.plot_data(selected_attr)

    def get_stats(self):
        if self.attr_data is None:
            return
        if not self.log_scale:
            data = self.attr_data
        else:
            data = np.log10(self.attr_data)
            # convert non real numbers to nan
            data = data.replace([np.inf, -np.inf], np.nan)
        # only real data
        data = data.dropna()
        self.data_to_plot = data
        self.mean = np.nanmean(data)
        self.std = np.nanstd(data)
        self.median = np.nanmedian(data)
        self.perc75 = np.nanpercentile(data, 75)
        self.perc25 = np.nanpercentile(data, 25)
        self.iqr = self.perc75 - self.perc25

    def draw_stats(self):
        if self.attr_data is None:
            return
        # draw lines for mean, median, std, percentiles on the canvas
        ax = self.canvas.figure.get_axes()[0]
        if self.is_median:
            ax.axvline(self.perc25, color='r', linestyle='--', label='25th percentile')
            ax.axvline(self.median, color='m', linestyle='-', label='Median')
            ax.axvline(self.perc75, color='r', linestyle='--', label='75th percentile')
        else:
            ax.axvline(self.mean - self.std, color='b', linestyle='--', label='Mean - Std')
            ax.axvline(self.mean, color='c', linestyle='-', label='Mean')
            ax.axvline(self.mean + self.std, color='b', linestyle='--', label='Mean + Std')
        ax.legend()
        self.canvas.draw()

    def plot_data(self, title):
        self.canvas.figure.clear()
        ax = self.canvas.figure.add_subplot(111)
        self.data_to_plot = self.data_to_plot.replace([np.inf, -np.inf], np.nan)
        try:
            if self.hist_reset:
                nbins = int(len(self.attr_data) ** 0.5)  # pretty nbins
                ax.hist(self.data_to_plot, bins=nbins)
                hist_min = ax.get_xlim()[0]
                hist_max = ax.get_xlim()[1]
            else:
                nbins = self.num_bins.value()
                hist_min = self.hist_min.value()
                hist_max = self.hist_max.value()
                ax.hist(self.data_to_plot, bins=nbins, range=(hist_min, hist_max))
        except ValueError:
            nbins = 10
            hist_min = 0
            hist_max = 1
        if self.is_median:
            full_title = f"{title}\n\nQuartiles: {self.perc25:.4f}, {self.median:.4f}, {self.perc75:.4f}"
        else:
            full_title = f"{title}\n\nMean: {self.mean:.4f}, Std: {self.std:.4f}"
        if self.match_t:
            full_title += f"\nTimepoint: {self.viewer.dims.current_step[0]}"
        else:
            full_title += f"\nTimepoint: all (pooled)"
        ax.set_title(full_title)
        if not self.log_scale:
            ax.set_xlabel("Value")
        else:
            ax.set_xlabel("Value (log10)")
        ax.set_ylabel("Frequency")
        self.canvas.draw()
        self.draw_stats()

        # if self.hist_min is not enabled
        if self.hist_reset:
            self.hist_min.setEnabled(True)
            self.hist_min.setValue(hist_min)
            self.hist_min.setRange(hist_min, hist_max)
            self.hist_min.setSingleStep((hist_max - hist_min) / 100)

            self.hist_max.setEnabled(True)
            self.hist_max.setValue(hist_max)
            self.hist_max.setRange(hist_min, hist_max)
            self.hist_max.setSingleStep((hist_max - hist_min) / 100)

            self.num_bins.setEnabled(True)
            self.num_bins.setValue(nbins)
            self.num_bins.setRange(1, len(self.attr_data))
            self.hist_reset = False


    def on_log_scale(self, state):
        self.hist_reset = True
        if state == 2:
            self.log_scale = True
        else:
            self.log_scale = False
        # self.plot_data(self.dropdown_attr.currentText())
        self.on_attr_selected(self.dropdown_attr.currentIndex())

if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



from itertools import combinations_with_replacement

from nellie import logger
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image, bbox
import numpy as np
from nellie import ndi, xp, device_type

from nellie.utils.gpu_functions import triangle_threshold, otsu_threshold


class Filter:
    def __init__(self, im_info: ImInfo,
                 num_t=None, remove_edges=True,
                 min_radius_um=0.20, max_radius_um=1, alpha_sq=0.5, beta_sq=0.5):
        self.im_info = im_info
        if not self.im_info.no_z:
            self.z_ratio = self.im_info.dim_sizes['Z'] / self.im_info.dim_sizes['X']
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]
        self.remove_edges = remove_edges
        # either (roughly) diffraction limit, or pixel size, whichever is larger
        self.min_radius_um = max(min_radius_um, self.im_info.dim_sizes['X'])
        self.max_radius_um = max_radius_um

        self.min_radius_px = self.min_radius_um / self.im_info.dim_sizes['X']
        self.max_radius_px = self.max_radius_um / self.im_info.dim_sizes['X']

        self.im_memmap = None
        self.frangi_memmap = None

        self.sigma_vec = None
        self.sigmas = None

        self.alpha_sq = alpha_sq
        self.beta_sq = beta_sq

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for frangi filter.')
        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)
        self.shape = self.im_memmap.shape
        im_frangi_path = self.im_info.pipeline_paths['im_frangi']
        self.frangi_memmap = self.im_info.allocate_memory(im_frangi_path, shape=self.shape, dtype='double',
                                                          description='frangi filtered im',
                                                          return_memmap=True)

    def _get_sigma_vec(self, sigma):
        if self.im_info.no_z:
            self.sigma_vec = (sigma, sigma)
        else:
            self.sigma_vec = (sigma / self.z_ratio, sigma, sigma)
        return self.sigma_vec

    def _set_default_sigmas(self):
        logger.debug('Setting to sigma values.')
        min_sigma_step_size = 0.2
        num_sigma = 5

        self.sigma_min = self.min_radius_px / 2
        self.sigma_max = self.max_radius_px / 3

        sigma_step_size_calculated = (self.sigma_max - self.sigma_min) / num_sigma
        sigma_step_size = max(min_sigma_step_size, sigma_step_size_calculated)  # Avoid taking too small of steps.

        self.sigmas = list(np.arange(self.sigma_min, self.sigma_max, sigma_step_size))
        logger.debug(f'Calculated sigma step size = {sigma_step_size_calculated}. Sigmas = {self.sigmas}')

    def _gauss_filter(self, sigma, t=None):
        self._get_sigma_vec(sigma)
        gauss_volume = xp.asarray(self.im_memmap[t, ...], dtype='double')
        logger.debug(f'Gaussian filtering {t=} with {self.sigma_vec=}.')

        gauss_volume = ndi.gaussian_filter(gauss_volume, sigma=self.sigma_vec,
                                           mode='reflect', cval=0.0, truncate=3).astype('double')
        return gauss_volume

    def _calculate_gamma(self, gauss_volume):
        gamma_tri = triangle_threshold(gauss_volume[gauss_volume > 0])
        gamma_otsu, _ = otsu_threshold(gauss_volume[gauss_volume > 0])
        gamma = min(gamma_tri, gamma_otsu)
        return gamma

    def _compute_hessian(self, image, mask=True):
        gradients = xp.gradient(image)
        axes = range(image.ndim)
        h_elems = xp.array([xp.gradient(gradients[ax0], axis=ax1).astype('float16')
                            for ax0, ax1 in combinations_with_replacement(axes, 2)])
        if mask:
            h_mask = self._get_frob_mask(h_elems)
        else:
            h_mask = xp.ones_like(image, dtype='bool')
        if self.remove_edges:
            h_mask = self._remove_edges(h_mask)

        if self.im_info.no_z:
            if device_type == 'cuda':
                hxx, hxy, hyy = [elem[..., np.newaxis, np.newaxis] for elem in h_elems[:, h_mask].get()]
            else:
                hxx, hxy, hyy = [elem[..., np.newaxis, np.newaxis] for elem in h_elems[:, h_mask]]
            hessian_matrices = np.concatenate([
                np.concatenate([hxx, hxy], axis=-1),
                np.concatenate([hxy, hyy], axis=-1)
            ], axis=-2)
        else:
            if device_type == 'cuda':
                hxx, hxy, hxz, hyy, hyz, hzz = [elem[..., np.newaxis, np.newaxis] for elem in h_elems[:, h_mask].get()]
            else:
                hxx, hxy, hxz, hyy, hyz, hzz = [elem[..., np.newaxis, np.newaxis] for elem in h_elems[:, h_mask]]
            hessian_matrices = np.concatenate([
                np.concatenate([hxx, hxy, hxz], axis=-1),
                np.concatenate([hxy, hyy, hyz], axis=-1),
                np.concatenate([hxz, hyz, hzz], axis=-1)
            ], axis=-2)

        return h_mask, hessian_matrices

    def _get_frob_mask(self, hessian_matrices):
        rescaled_hessian = hessian_matrices / xp.max(xp.abs(hessian_matrices))
        frobenius_norm = xp.linalg.norm(rescaled_hessian, axis=0)
        frobenius_norm[xp.isinf(frobenius_norm)] = xp.max(frobenius_norm[~xp.isinf(frobenius_norm)])
        non_zero_frobenius = frobenius_norm[frobenius_norm > 0]
        if len(non_zero_frobenius) == 0:
            frobenius_threshold = 0
        else:
            frob_triangle_thresh = triangle_threshold(non_zero_frobenius)
            frob_otsu_thresh, _ = otsu_threshold(non_zero_frobenius)
            frobenius_threshold = min(frob_triangle_thresh, frob_otsu_thresh)
        mask = frobenius_norm > frobenius_threshold
        return mask

    def _compute_chunkwise_eigenvalues(self, hessian_matrices, chunk_size=1E6):
        chunk_size = int(chunk_size)
        total_voxels = len(hessian_matrices)

        eigenvalues_list = []

        if chunk_size is None:  # chunk size is entire vector
            chunk_size = total_voxels

        # iterate over chunks
        for start_idx in range(0, total_voxels, int(chunk_size)):
            end_idx = min(start_idx + chunk_size, total_voxels)
            gpu_chunk = xp.array(hessian_matrices[start_idx:end_idx])
            chunk_eigenvalues = xp.linalg.eigvalsh(gpu_chunk)
            eigenvalues_list.append(chunk_eigenvalues)

        # concatenate all the eigval chunks and reshape to the original spatial structure
        eigenvalues_flat = xp.concatenate(eigenvalues_list, axis=0)
        sort_order = xp.argsort(xp.abs(eigenvalues_flat), axis=1)
        eigenvalues_flat = xp.take_along_axis(eigenvalues_flat, sort_order, axis=1)

        return eigenvalues_flat

    def _filter_hessian(self, eigenvalues, gamma_sq):
        if self.im_info.no_z:
            rb_sq = (xp.abs(eigenvalues[:, 0]) / xp.abs(eigenvalues[:, 1])) ** 2
            s_sq = (eigenvalues[:, 0] ** 2) + (eigenvalues[:, 1] ** 2)
            filtered_im = (xp.exp(-(rb_sq / self.beta_sq))) * (1 - xp.exp(-(s_sq / gamma_sq)))
        else:
            ra_sq = (xp.abs(eigenvalues[:, 1]) / xp.abs(eigenvalues[:, 2])) ** 2
            rb_sq = (xp.abs(eigenvalues[:, 1]) / xp.sqrt(xp.abs(eigenvalues[:, 1] * eigenvalues[:, 2]))) ** 2
            s_sq = (xp.sqrt((eigenvalues[:, 0] ** 2) + (eigenvalues[:, 1] ** 2) + (eigenvalues[:, 2] ** 2))) ** 2
            filtered_im = (1 - xp.exp(-(ra_sq / self.alpha_sq))) * (xp.exp(-(rb_sq / self.beta_sq))) * \
                          (1 - xp.exp(-(s_sq / gamma_sq)))
        if not self.im_info.no_z:
            filtered_im[eigenvalues[:, 2] > 0] = 0
        filtered_im[eigenvalues[:, 1] > 0] = 0
        filtered_im = xp.nan_to_num(filtered_im, False, 1)
        return filtered_im

    def _filter_log(self, frame, mask):
        lapofg = xp.zeros_like(frame, dtype='double')
        for i, s in enumerate(self.sigmas):
            sigma_vec = self._get_sigma_vec(s)
            current_lapofg = -ndi.gaussian_laplace(frame, sigma_vec) * xp.mean(s) ** 2
            current_lapofg = current_lapofg * mask
            min_indices = current_lapofg < lapofg
            lapofg[min_indices] = current_lapofg[min_indices]
            if i == 0:
                lapofg = current_lapofg
        lapofg_min_proj = lapofg
        return lapofg_min_proj

    def _run_frame(self, t, mask=True):
        logger.info(f'Running frangi filter on {t=}.')
        vesselness = xp.zeros_like(self.im_memmap[t, ...], dtype='float64')
        temp = xp.zeros_like(self.im_memmap[t, ...], dtype='float64')
        masks = xp.ones_like(self.im_memmap[t, ...], dtype='bool')
        for sigma_num, sigma in enumerate(self.sigmas):
            gauss_volume = self._gauss_filter(sigma, t)  # * xp.mean(sigma) ** 2

            gamma = self._calculate_gamma(gauss_volume)
            gamma_sq = 2 * gamma ** 2

            h_mask, hessian_matrices = self._compute_hessian(gauss_volume, mask=mask)
            if len(hessian_matrices) == 0:
                continue
            eigenvalues = self._compute_chunkwise_eigenvalues(hessian_matrices.astype('float'))

            temp[h_mask] = self._filter_hessian(eigenvalues, gamma_sq=gamma_sq)

            max_indices = temp > vesselness
            vesselness[max_indices] = temp[max_indices]
            masks = xp.where(~h_mask, 0, masks)

        vesselness = vesselness * masks
        return vesselness

    def _mask_volume(self, frangi_frame):
        frangi_threshold = xp.percentile(frangi_frame[frangi_frame > 0], 1)
        frangi_mask = frangi_frame > frangi_threshold
        frangi_mask = ndi.binary_opening(frangi_mask)
        frangi_frame = frangi_frame * frangi_mask
        return frangi_frame

    def _remove_edges(self, frangi_frame):
        if self.im_info.no_z:
            num_z = 1
        else:
            num_z = self.im_info.shape[self.im_info.axes.index('Z')]
        for z_idx in range(num_z):
            if self.im_info.no_z:
                rmin, rmax, cmin, cmax = bbox(frangi_frame)
            else:
                rmin, rmax, cmin, cmax = bbox(frangi_frame[z_idx, ...])
            frangi_frame[z_idx, rmin:rmin + 15, ...] = 0
            frangi_frame[z_idx, rmax - 15:rmax + 1, ...] = 0
        return frangi_frame

    def _run_filter(self, mask=True):
        for t in range(self.num_t):
            frangi_frame = self._run_frame(t, mask=mask)
            if not xp.sum(frangi_frame):
                frangi_frame = self._mask_volume(frangi_frame)
            filtered_im = frangi_frame

            if device_type == 'cuda':
                filtered_im = filtered_im.get()

            if self.im_info.no_t or self.num_t == 1:
                self.frangi_memmap[:] = filtered_im[:]
            else:
                self.frangi_memmap[t, ...] = filtered_im
            self.frangi_memmap.flush()

    def run(self, mask=True):
        logger.info('Running frangi filter.')
        self._get_t()
        self._allocate_memory()
        self._set_default_sigmas()
        self._run_filter(mask=mask)


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_gav_tests\fibro_3.nd2"
    im_info = ImInfo(im_path)
    filter_im = Filter(im_info, num_t=2)
    filter_im.run()



import pickle

import numpy as np
from scipy import spatial
from skimage.measure import regionprops

from nellie import logger
from nellie.im_info.im_info import ImInfo
from nellie.tracking.flow_interpolation import FlowInterpolator
from nellie.utils.general import get_reshaped_image
import pandas as pd
import time


class Hierarchy:
    def __init__(self, im_info: ImInfo, num_t: int):
        self.im_info = im_info
        self.num_t = num_t
        if self.im_info.no_z:
            self.spacing = (self.im_info.dim_sizes['Y'], self.im_info.dim_sizes['X'])
        else:
            self.spacing = (self.im_info.dim_sizes['Z'], self.im_info.dim_sizes['Y'], self.im_info.dim_sizes['X'])

        self.im_raw = None
        self.im_struct = None
        self.im_distance = None
        self.im_skel = None
        self.im_pixel_class = None
        self.label_components = None
        self.label_branches = None
        self.im_border_mask = None

        self.flow_interpolator_fw = FlowInterpolator(im_info)
        self.flow_interpolator_bw = FlowInterpolator(im_info, forward=False)

        self.shape = None

        self.voxels = None
        self.nodes = None
        self.branches = None
        self.components = None
        self.image = None

    def _get_t(self):
        if self.num_t is None and not self.im_info.no_t:
            # if self.im_info.no_t:
            #     raise ValueError("No time dimension in image.")
            self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        # if self.num_t < 3:
        #     raise ValueError("num_t must be at least 3")
        return self.num_t

    def _allocate_memory(self):
        im_raw = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_raw = get_reshaped_image(im_raw, self.num_t, self.im_info)
        im_struct = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_frangi'])
        self.im_struct = get_reshaped_image(im_struct, self.num_t, self.im_info)
        label_components = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_instance_label'])
        self.label_components = get_reshaped_image(label_components, self.num_t, self.im_info)
        label_branches = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_skel_relabelled'])
        self.label_branches = get_reshaped_image(label_branches, self.num_t, self.im_info)
        im_skel = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_skel'])
        self.im_skel = get_reshaped_image(im_skel, self.num_t, self.im_info)
        im_pixel_class = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_pixel_class'])
        self.im_pixel_class = get_reshaped_image(im_pixel_class, self.num_t, self.im_info)
        im_distance = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_distance'])
        self.im_distance = get_reshaped_image(im_distance, self.num_t, self.im_info)
        im_border_mask = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_border'])
        self.im_border_mask = get_reshaped_image(im_border_mask, self.num_t, self.im_info)
        if not self.im_info.no_t:
            im_obj_reassigned = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_obj_label_reassigned'])
            self.im_obj_reassigned = get_reshaped_image(im_obj_reassigned, self.num_t, self.im_info)
            im_branch_reassigned = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_branch_label_reassigned'])
            self.im_branch_reassigned = get_reshaped_image(im_branch_reassigned, self.num_t, self.im_info)

        self.shape = self.im_raw.shape
        # self.im_info.shape = self.shape

    def _get_hierarchies(self):
        self.voxels = Voxels(self)
        logger.info("Running voxel analysis")
        start = time.time()
        self.voxels.run()
        end = time.time()
        v_time = end - start

        self.nodes = Nodes(self)
        logger.info("Running node analysis")
        start = time.time()
        self.nodes.run()
        end = time.time()
        n_time = end - start

        self.branches = Branches(self)
        logger.info("Running branch analysis")
        start = time.time()
        self.branches.run()
        end = time.time()
        b_time = end - start

        self.components = Components(self)
        logger.info("Running component analysis")
        start = time.time()
        self.components.run()
        end = time.time()
        c_time = end - start

        self.image = Image(self)
        logger.info("Running image analysis")
        start = time.time()
        self.image.run()
        end = time.time()
        i_time = end - start

        logger.debug(f"Voxel analysis took {v_time} seconds")
        logger.debug(f"Node analysis took {n_time} seconds")
        logger.debug(f"Branch analysis took {b_time} seconds")
        logger.debug(f"Component analysis took {c_time} seconds")
        logger.debug(f"Image analysis took {i_time} seconds")

    def _save_dfs(self):
        voxel_features, voxel_headers = create_feature_array(self.voxels)
        voxel_df = pd.DataFrame(voxel_features, columns=voxel_headers)
        voxel_df.to_csv(self.im_info.pipeline_paths['features_voxels'], index=True)

        node_features, node_headers = create_feature_array(self.nodes)
        node_df = pd.DataFrame(node_features, columns=node_headers)
        node_df.to_csv(self.im_info.pipeline_paths['features_nodes'], index=True)

        branch_features, branch_headers = create_feature_array(self.branches, self.branches.branch_label)
        branch_df = pd.DataFrame(branch_features, columns=branch_headers)
        branch_df.to_csv(self.im_info.pipeline_paths['features_branches'], index=True)

        component_features, component_headers = create_feature_array(self.components, self.components.component_label)
        component_df = pd.DataFrame(component_features, columns=component_headers)
        component_df.to_csv(self.im_info.pipeline_paths['features_components'], index=True)

        image_features, image_headers = create_feature_array(self.image)
        image_df = pd.DataFrame(image_features, columns=image_headers)
        image_df.to_csv(self.im_info.pipeline_paths['features_image'], index=True)

    def _save_adjacency_maps(self):
        # edge list:
        v_b = [self.voxels.branch_labels[t][:, None] == self.branches.branch_label[t] for t in
               range(len(self.voxels.time))]
        v_n = []
        for t in range(len(self.voxels.time)):
            num_voxels = len(self.voxels.coords[t])
            num_nodes = len(self.nodes.nodes[t])
            v_n_temp = np.zeros((num_voxels, num_nodes), dtype=bool)
            for voxel, nodes in enumerate(self.voxels.node_labels[t]):
                v_n_temp[voxel, nodes] = True
            v_n.append(v_n_temp)
        v_o = [self.voxels.component_labels[t][:, None] == self.components.component_label[t] for t in
               range(len(self.voxels.time))]
        v_i = [np.ones((len(self.voxels.coords[t]), 1), dtype=bool) for t in range(len(self.voxels.time))]

        n_v = [v_n[t].T for t in range(len(self.voxels.time))]
        n_b = [self.nodes.branch_label[t][:, None] == self.branches.branch_label[t] for t in
               range(len(self.nodes.time))]
        n_o = [self.nodes.component_label[t][:, None] == self.components.component_label[t] for t in
               range(len(self.nodes.time))]
        n_i = [np.ones((len(self.nodes.nodes[t]), 1), dtype=bool) for t in range(len(self.nodes.time))]

        b_v = [v_b[t].T for t in range(len(self.voxels.time))]
        b_n = [n_b[t].T for t in range(len(self.nodes.time))]
        b_o = [self.branches.component_label[t][:, None] == self.components.component_label[t] for t in
               range(len(self.branches.time))]
        b_i = [np.ones((len(self.branches.branch_label[t]), 1), dtype=bool) for t in
               range(len(self.branches.time))]

        o_v = [v_o[t].T for t in range(len(self.voxels.time))]
        o_n = [n_o[t].T for t in range(len(self.nodes.time))]
        o_b = [b_o[t].T for t in range(len(self.branches.time))]
        o_i = [np.ones((len(self.components.component_label[t]), 1), dtype=bool) for t in
               range(len(self.components.time))]

        i_v = [v_i[t].T for t in range(len(self.voxels.time))]
        i_n = [n_i[t].T for t in range(len(self.nodes.time))]
        i_b = [b_i[t].T for t in range(len(self.branches.time))]
        i_o = [o_i[t].T for t in range(len(self.components.time))]

        # create a dict with all the edges
        # could also link voxels between t frames
        edges = {
            "v_b": v_b, "v_n": v_n, "v_o": v_o, "v_i": v_i,
            "n_v": n_v, "n_b": n_b, "n_o": n_o, "n_i": n_i,
            "b_v": b_v, "b_n": b_n, "b_o": b_o, "b_i": b_i,
            "o_v": o_v, "o_n": o_n, "o_b": o_b, "o_i": o_i,
            "i_v": i_v, "i_n": i_n, "i_b": i_b, "i_o": i_o,
        }
        # pickle and save
        with open(self.im_info.pipeline_paths['adjacency_maps'], "wb") as f:
            pickle.dump(edges, f)

    def run(self):
        self._get_t()
        self._allocate_memory()
        self._get_hierarchies()
        print("Saving dataframes and adjacency maps")
        self._save_dfs()
        self._save_adjacency_maps()


def append_to_array(to_append):
    new_array = []
    new_headers = []
    for feature, stats in to_append.items():
        if type(stats) is not dict:
            stats = {'raw': stats}
        for stat, vals in stats.items():
            vals = np.array(vals)
            if len(vals.shape) > 1:
                for i in range(len(vals[0])):
                    new_array.append(vals[:, i])
                    new_headers.append(f'{feature}_{stat}_{i}')
            else:
                new_array.append(vals)
                new_headers.append(f'{feature}_{stat}')
    return new_array, new_headers


def create_feature_array(level, labels=None):
    full_array = None
    headers = None
    all_attr = []
    attr_dict = []
    if node_attr := getattr(level, 'aggregate_node_metrics', None):
        all_attr.append(node_attr)
    if voxel_attr := getattr(level, 'aggregate_voxel_metrics', None):
        all_attr.append(voxel_attr)
    if branch_attr := getattr(level, 'aggregate_branch_metrics', None):
        all_attr.append(branch_attr)
    if component_attr := getattr(level, 'aggregate_component_metrics', None):
        all_attr.append(component_attr)
    inherent_features = level.stats_to_aggregate
    for feature in inherent_features:
        if feature_vals := getattr(level, feature, None):
            all_attr.append([{feature: feature_vals[t]} for t in range(len(feature_vals))])

    for t in range(len(all_attr[0])):
        time_dict = {}
        for attr in all_attr:
            time_dict.update(attr[t])
        attr_dict.append(time_dict)

    for t in range(len(attr_dict)):
        to_append = attr_dict[t]
        time_array, new_headers = append_to_array(to_append)
        if labels is None:
            labels_t = range(len(time_array[0]))
        else:
            labels_t = labels[t]
        time_array.insert(0, labels_t)
        # append a list of t values to the start of time_array
        time_array.insert(0, np.array([t] * len(time_array[0])))
        if headers is None:
            headers = new_headers
        if full_array is None:
            full_array = np.array(time_array).T
        else:
            time_array = np.array(time_array).T
            full_array = np.vstack([full_array, time_array])

    headers.insert(0, 'label')
    headers.insert(0, 't')
    return full_array, headers


class Voxels:
    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.coords = []

        # add voxel metrics
        self.intensity = []
        self.structure = []

        self.vec01 = []
        self.vec12 = []

        self.ang_acc = []
        self.ang_acc_mag = []
        self.ang_vel_mag = []
        self.ang_vel_orient = []
        self.ang_vel = []
        self.lin_acc = []
        self.lin_acc_mag = []
        self.lin_vel_mag = []
        self.lin_vel_orient = []
        self.lin_vel = []

        self.ang_acc_rel = []
        self.ang_acc_rel_mag = []
        self.ang_vel_mag_rel = []
        self.ang_vel_orient_rel = []
        self.ang_vel_rel = []
        self.lin_acc_rel = []
        self.lin_acc_rel_mag = []
        self.lin_vel_mag_rel = []
        self.lin_vel_orient_rel = []
        self.lin_vel_rel = []
        self.directionality_rel = []
        self.directionality_acc_rel = []

        self.ang_acc_com = []
        self.ang_acc_com_mag = []
        self.ang_vel_mag_com = []
        self.ang_vel_orient_com = []
        self.ang_vel_com = []
        self.lin_acc_com = []
        self.lin_acc_com_mag = []
        self.lin_vel_mag_com = []
        self.lin_vel_orient_com = []
        self.lin_vel_com = []
        self.directionality_com = []
        self.directionality_acc_com = []

        self.node_labels = []
        self.branch_labels = []
        self.component_labels = []
        self.image_name = []

        self.node_dim0_lims = []
        self.node_dim1_lims = []
        self.node_dim2_lims = []
        self.node_voxel_idxs = []

        self.stats_to_aggregate = [
            "ang_acc", "ang_acc_com", "ang_acc_com_mag", "ang_acc_mag", "ang_acc_rel", "ang_acc_rel_mag",
            "ang_vel", "ang_vel_com", "ang_vel_mag", "ang_vel_mag_com", "ang_vel_mag_rel",
            "ang_vel_orient", "ang_vel_orient_com", "ang_vel_orient_rel", "ang_vel_rel",
            "directionality_acc_com", "directionality_acc_rel", "directionality_com", "directionality_rel",
            "lin_acc", "lin_acc_com", "lin_acc_com_mag", "lin_acc_mag", "lin_acc_rel", "lin_acc_rel_mag",
            "lin_vel", "lin_vel_com", "lin_vel_mag", "lin_vel_mag_rel", "lin_vel_orient", "lin_vel_orient_com",
            "lin_vel_orient_rel", "lin_vel_mag_com",
            "lin_vel_rel", "intensity", "structure"
        ]

        self.features_to_save = self.stats_to_aggregate

    def _get_node_info(self, t, frame_coords):
        # get all network pixels
        skeleton_pixels = np.argwhere(self.hierarchy.im_pixel_class[t] > 0)
        skeleton_radius = self.hierarchy.im_distance[t][tuple(skeleton_pixels.T)]

        # create bounding boxes of size largest_thickness around each skeleton pixel
        lims_dim0 = (skeleton_radius[:, np.newaxis] * np.array([-1, 1]) + skeleton_pixels[:, 0, np.newaxis]).astype(int)
        lims_dim0[:, 1] += 1
        lims_dim1 = (skeleton_radius[:, np.newaxis] * np.array([-1, 1]) + skeleton_pixels[:, 1, np.newaxis]).astype(int)
        lims_dim1[:, 1] += 1

        lims_dim0[lims_dim0 < 0] = 0
        lims_dim1[lims_dim1 < 0] = 0

        if not self.hierarchy.im_info.no_z:
            lims_dim2 = (skeleton_radius[:, np.newaxis] * np.array([-1, 1]) + skeleton_pixels[:, 2, np.newaxis]).astype(
                int)
            lims_dim2[:, 1] += 1
            lims_dim2[lims_dim2 < 0] = 0
            max_dim0 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index('Z')]
            max_dim1 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index('Y')]
            max_dim2 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index('X')]
            lims_dim2[lims_dim2 > max_dim2] = max_dim2
        else:
            max_dim0 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index('Y')]
            max_dim1 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index('X')]

        lims_dim0[lims_dim0 > max_dim0] = max_dim0
        lims_dim1[lims_dim1 > max_dim1] = max_dim1

        self.node_dim0_lims.append(lims_dim0)
        self.node_dim1_lims.append(lims_dim1)

        self.node_dim2_lims.append(lims_dim2) if not self.hierarchy.im_info.no_z else None

        frame_coord_nodes_idxs = []
        # if a frame coord is within the bounding box of a skeleton pixel, add the skeleton pixel's
        #  idx to the list for that frame coord
        for i, frame_coord in enumerate(frame_coords):
            if not self.hierarchy.im_info.no_z:
                dim0_coord, dim1_coord, dim2_coord = frame_coord
                dim2_mask = (lims_dim2[:, 0] <= dim2_coord) & (lims_dim2[:, 1] >= dim2_coord)
            else:
                dim0_coord, dim1_coord = frame_coord
            dim0_mask = (lims_dim0[:, 0] <= dim0_coord) & (lims_dim0[:, 1] >= dim0_coord)
            dim1_mask = (lims_dim1[:, 0] <= dim1_coord) & (lims_dim1[:, 1] >= dim1_coord)
            mask = dim0_mask & dim1_mask & dim2_mask if not self.hierarchy.im_info.no_z else dim0_mask & dim1_mask
            frame_coord_nodes_idxs.append(np.argwhere(mask).flatten())

        self.node_labels.append(frame_coord_nodes_idxs)

        # initialize an empty dictionary for frame_coord_nodes_idxs
        node_voxel_idxs = []

        for i, skeleton_px in enumerate(skeleton_pixels):
            if not self.hierarchy.im_info.no_z:
                bbox_voxels = np.argwhere(
                    (lims_dim0[i, 0] <= frame_coords[:, 0]) & (lims_dim0[i, 1] >= frame_coords[:, 0]) &
                    (lims_dim1[i, 0] <= frame_coords[:, 1]) & (lims_dim1[i, 1] >= frame_coords[:, 1]) &
                    (lims_dim2[i, 0] <= frame_coords[:, 2]) & (lims_dim2[i, 1] >= frame_coords[:, 2]))
            else:
                bbox_voxels = np.argwhere(
                    (lims_dim0[i, 0] <= frame_coords[:, 0]) & (lims_dim0[i, 1] >= frame_coords[:, 0]) &
                    (lims_dim1[i, 0] <= frame_coords[:, 1]) & (lims_dim1[i, 1] >= frame_coords[:, 1]))
            node_voxel_idxs.append(bbox_voxels.flatten())

        self.node_voxel_idxs.append(node_voxel_idxs)

    def _get_min_euc_dist(self, t, vec):
        euc_dist = np.linalg.norm(vec, axis=1)
        branch_labels = self.branch_labels[t]

        df = pd.DataFrame({'euc_dist': euc_dist, 'branch_label': branch_labels})
        df = df[~np.isnan(df['euc_dist'])]
        idxmin = df.groupby('branch_label')['euc_dist'].idxmin()
        # if there are no non-nan values in a branch, give idxmin at that branch index a value of nan
        missing_branches = np.setdiff1d(np.unique(branch_labels), df['branch_label'])
        for missing_branch in missing_branches:
            idxmin[missing_branch] = np.nan

        return idxmin

    def _get_ref_coords(self, coords_a, coords_b, idxmin, t):
        vals_a = idxmin[self.branch_labels[t]].values
        vals_a_no_nan = vals_a.copy()
        vals_a_no_nan[np.isnan(vals_a_no_nan)] = 0
        vals_a_no_nan = vals_a_no_nan.astype(int)

        vals_b = idxmin[self.branch_labels[t]].values
        vals_b_no_nan = vals_b.copy()
        vals_b_no_nan[np.isnan(vals_b_no_nan)] = 0
        vals_b_no_nan = vals_b_no_nan.astype(int)
        ref_a = coords_a[vals_a_no_nan]
        ref_b = coords_b[vals_b_no_nan]

        ref_a[np.isnan(vals_a)] = np.nan
        ref_b[np.isnan(vals_b)] = np.nan

        return ref_a, ref_b

    def _get_motility_stats(self, t, coords_1_px):
        coords_1_px = coords_1_px.astype('float32')
        if self.hierarchy.im_info.no_z:
            dims = 2
        else:
            dims = 3

        vec01 = []
        vec12 = []
        if t > 0:
            vec01_px = self.hierarchy.flow_interpolator_bw.interpolate_coord(coords_1_px, t)
            vec01 = vec01_px * self.hierarchy.spacing
            self.vec01.append(vec01)
        else:
            self.vec01.append(np.full((len(coords_1_px), dims), np.nan))

        if t < self.hierarchy.num_t - 1:
            vec12_px = self.hierarchy.flow_interpolator_fw.interpolate_coord(coords_1_px, t)
            vec12 = vec12_px * self.hierarchy.spacing
            self.vec12.append(vec12)
        else:
            self.vec12.append(np.full((len(coords_1_px), dims), np.nan))

        coords_1 = coords_1_px * self.hierarchy.spacing
        coords_com_1 = np.nanmean(coords_1, axis=0)
        r1_rel_com = coords_1 - coords_com_1
        r1_com_mag = np.linalg.norm(r1_rel_com, axis=1)

        if len(vec01) and len(vec12):
            coords_0_px = coords_1_px - vec01_px
            coords_0 = coords_0_px * self.hierarchy.spacing

            lin_vel_01, lin_vel_mag_01, lin_vel_orient_01 = self._get_linear_velocity(coords_0, coords_1)
            ang_vel_01, ang_vel_mag_01, ang_vel_orient_01 = self._get_angular_velocity(coords_0, coords_1)

            idxmin01 = self._get_min_euc_dist(t, vec01)
            ref_coords01 = self._get_ref_coords(coords_0, coords_1, idxmin01, t)
            ref_coords01[0][np.isnan(vec01)] = np.nan
            ref_coords01[1][np.isnan(vec01)] = np.nan
            r0_rel_01 = coords_0 - ref_coords01[0]
            r1_rel_01 = coords_1 - ref_coords01[1]

            lin_vel_rel_01, lin_vel_mag_rel_01, lin_vel_orient_rel_01 = self._get_linear_velocity(r0_rel_01, r1_rel_01)
            ang_vel_rel_01, ang_vel_mag_rel_01, ang_vel_orient_rel_01 = self._get_angular_velocity(r0_rel_01, r1_rel_01)

            coords_com_0 = np.nanmean(coords_0, axis=0)
            r0_rel_com = coords_0 - coords_com_0
            lin_vel_com_01, lin_vel_mag_com_01, lin_vel_orient_com_01 = self._get_linear_velocity(r0_rel_com,
                                                                                                  r1_rel_com)
            ang_vel_com_01, ang_vel_mag_com_01, ang_vel_orient_com_01 = self._get_angular_velocity(r0_rel_com,
                                                                                                   r1_rel_com)

            r0_com_mag = np.linalg.norm(r0_rel_com, axis=1)
            directionality_com_01 = np.abs(r0_com_mag - r1_com_mag) / (r0_com_mag + r1_com_mag)

            r0_rel_mag_01 = np.linalg.norm(r0_rel_01, axis=1)
            r1_rel_mag_01 = np.linalg.norm(r1_rel_01, axis=1)
            directionality_rel_01 = np.abs(r0_rel_mag_01 - r1_rel_mag_01) / (r0_rel_mag_01 + r1_rel_mag_01)

        if len(vec12):
            coords_2_px = coords_1_px + vec12_px
            coords_2 = coords_2_px * self.hierarchy.spacing

            lin_vel, lin_vel_mag, lin_vel_orient = self._get_linear_velocity(coords_1, coords_2)
            ang_vel, ang_vel_mag, ang_vel_orient = self._get_angular_velocity(coords_1, coords_2)

            idxmin12 = self._get_min_euc_dist(t, vec12)
            ref_coords12 = self._get_ref_coords(coords_1, coords_2, idxmin12, t)
            ref_coords12[0][np.isnan(vec12)] = np.nan
            ref_coords12[1][np.isnan(vec12)] = np.nan
            r1_rel_12 = coords_1 - ref_coords12[0]
            r2_rel_12 = coords_2 - ref_coords12[1]

            lin_vel_rel, lin_vel_mag_rel, lin_vel_orient_rel = self._get_linear_velocity(r1_rel_12, r2_rel_12)
            ang_vel_rel, ang_vel_mag_rel, ang_vel_orient_rel = self._get_angular_velocity(r1_rel_12, r2_rel_12)

            coords_com_2 = np.nanmean(coords_2, axis=0)
            r2_rel_com = coords_2 - coords_com_2
            lin_vel_com, lin_vel_mag_com, lin_vel_orient_com = self._get_linear_velocity(r1_rel_com, r2_rel_com)
            ang_vel_com, ang_vel_mag_com, ang_vel_orient_com = self._get_angular_velocity(r1_rel_com, r2_rel_com)

            r2_com_mag = np.linalg.norm(r2_rel_com, axis=1)
            directionality_com = np.abs(r2_com_mag - r1_com_mag) / (r2_com_mag + r1_com_mag)

            r2_rel_mag_12 = np.linalg.norm(r2_rel_12, axis=1)
            r1_rel_mag_12 = np.linalg.norm(r1_rel_12, axis=1)
            directionality_rel = np.abs(r2_rel_mag_12 - r1_rel_mag_12) / (r2_rel_mag_12 + r1_rel_mag_12)
        else:
            # vectors of nans
            lin_vel = np.full((len(coords_1), dims), np.nan)
            lin_vel_mag = np.full(len(coords_1), np.nan)
            lin_vel_orient = np.full((len(coords_1), dims), np.nan)
            ang_vel_mag = np.full(len(coords_1), np.nan)
            lin_vel_rel = np.full((len(coords_1), dims), np.nan)
            lin_vel_mag_rel = np.full(len(coords_1), np.nan)
            lin_vel_orient_rel = np.full((len(coords_1), dims), np.nan)
            ang_vel_mag_rel = np.full(len(coords_1), np.nan)
            lin_vel_com = np.full((len(coords_1), dims), np.nan)
            lin_vel_mag_com = np.full(len(coords_1), np.nan)
            lin_vel_orient_com = np.full((len(coords_1), dims), np.nan)
            ang_vel_mag_com = np.full(len(coords_1), np.nan)
            directionality_com = np.full(len(coords_1), np.nan)
            directionality_rel = np.full(len(coords_1), np.nan)
            if dims == 3:
                ang_vel = np.full((len(coords_1), dims), np.nan)
                ang_vel_orient = np.full((len(coords_1), dims), np.nan)
                ang_vel_rel = np.full((len(coords_1), dims), np.nan)
                ang_vel_orient_rel = np.full((len(coords_1), dims), np.nan)
                ang_vel_com = np.full((len(coords_1), dims), np.nan)
                ang_vel_orient_com = np.full((len(coords_1), dims), np.nan)
            else:
                ang_vel = np.full(len(coords_1), np.nan)
                ang_vel_orient = np.full(len(coords_1), np.nan)
                ang_vel_rel = np.full(len(coords_1), np.nan)
                ang_vel_orient_rel = np.full(len(coords_1), np.nan)
                ang_vel_com = np.full(len(coords_1), np.nan)
                ang_vel_orient_com = np.full(len(coords_1), np.nan)
        self.lin_vel.append(lin_vel)
        self.lin_vel_mag.append(lin_vel_mag)
        self.lin_vel_orient.append(lin_vel_orient)
        self.ang_vel.append(ang_vel)
        self.ang_vel_mag.append(ang_vel_mag)
        self.ang_vel_orient.append(ang_vel_orient)
        self.lin_vel_rel.append(lin_vel_rel)
        self.lin_vel_mag_rel.append(lin_vel_mag_rel)
        self.lin_vel_orient_rel.append(lin_vel_orient_rel)
        self.ang_vel_rel.append(ang_vel_rel)
        self.ang_vel_mag_rel.append(ang_vel_mag_rel)
        self.ang_vel_orient_rel.append(ang_vel_orient_rel)
        self.lin_vel_com.append(lin_vel_com)
        self.lin_vel_mag_com.append(lin_vel_mag_com)
        self.lin_vel_orient_com.append(lin_vel_orient_com)
        self.ang_vel_com.append(ang_vel_com)
        self.ang_vel_mag_com.append(ang_vel_mag_com)
        self.ang_vel_orient_com.append(ang_vel_orient_com)
        self.directionality_com.append(directionality_com)
        self.directionality_rel.append(directionality_rel)

        if len(vec01) and len(vec12):
            lin_acc = (lin_vel - lin_vel_01) / self.hierarchy.im_info.dim_sizes['T']
            lin_acc_mag = np.linalg.norm(lin_acc, axis=1)
            ang_acc = (ang_vel - ang_vel_01) / self.hierarchy.im_info.dim_sizes['T']
            lin_acc_rel = (lin_vel_rel - lin_vel_rel_01) / self.hierarchy.im_info.dim_sizes['T']
            lin_acc_rel_mag = np.linalg.norm(lin_acc_rel, axis=1)
            ang_acc_rel = (ang_vel_rel - ang_vel_rel_01) / self.hierarchy.im_info.dim_sizes['T']
            lin_acc_com = (lin_vel_com - lin_vel_com_01) / self.hierarchy.im_info.dim_sizes['T']
            lin_acc_com_mag = np.linalg.norm(lin_acc_com, axis=1)
            ang_acc_com = (ang_vel_com - ang_vel_com_01) / self.hierarchy.im_info.dim_sizes['T']
            if self.hierarchy.im_info.no_z:
                ang_acc_mag = np.abs(ang_acc)
                ang_acc_rel_mag = np.abs(ang_acc_rel)
                ang_acc_com_mag = np.abs(ang_acc_com)
            else:
                ang_acc_mag = np.linalg.norm(ang_acc, axis=1)
                ang_acc_rel_mag = np.linalg.norm(ang_acc_rel, axis=1)
                ang_acc_com_mag = np.linalg.norm(ang_acc_com, axis=1)
            # directionality acceleration is the change of directionality based on
            #  directionality_com_01 and directionality_com
            directionality_acc_com = np.abs(directionality_com - directionality_com_01)
            directionality_acc_rel = np.abs(directionality_rel - directionality_rel_01)
        else:
            # vectors of nans
            lin_acc = np.full((len(coords_1), dims), np.nan)
            lin_acc_mag = np.full(len(coords_1), np.nan)
            ang_acc_mag = np.full(len(coords_1), np.nan)
            lin_acc_rel = np.full((len(coords_1), dims), np.nan)
            lin_acc_rel_mag = np.full(len(coords_1), np.nan)
            ang_acc_rel_mag = np.full(len(coords_1), np.nan)
            lin_acc_com = np.full((len(coords_1), dims), np.nan)
            lin_acc_com_mag = np.full(len(coords_1), np.nan)
            ang_acc_com_mag = np.full(len(coords_1), np.nan)
            directionality_acc_com = np.full(len(coords_1), np.nan)
            directionality_acc_rel = np.full(len(coords_1), np.nan)
            if dims == 3:
                ang_acc = np.full((len(coords_1), dims), np.nan)
                ang_acc_rel = np.full((len(coords_1), dims), np.nan)
                ang_acc_com = np.full((len(coords_1), dims), np.nan)
            else:
                ang_acc = np.full(len(coords_1), np.nan)
                ang_acc_rel = np.full(len(coords_1), np.nan)
                ang_acc_com = np.full(len(coords_1), np.nan)
        self.lin_acc.append(lin_acc)
        self.lin_acc_mag.append(lin_acc_mag)
        self.ang_acc.append(ang_acc)
        self.ang_acc_mag.append(ang_acc_mag)
        self.lin_acc_rel.append(lin_acc_rel)
        self.lin_acc_rel_mag.append(lin_acc_rel_mag)
        self.ang_acc_rel.append(ang_acc_rel)
        self.ang_acc_rel_mag.append(ang_acc_rel_mag)
        self.lin_acc_com.append(lin_acc_com)
        self.lin_acc_com_mag.append(lin_acc_com_mag)
        self.ang_acc_com.append(ang_acc_com)
        self.ang_acc_com_mag.append(ang_acc_com_mag)
        self.directionality_acc_com.append(directionality_acc_com)
        self.directionality_acc_rel.append(directionality_acc_rel)

    def _get_linear_velocity(self, ra, rb):
        lin_disp = rb - ra
        lin_vel = lin_disp / self.hierarchy.im_info.dim_sizes['T']
        lin_vel_mag = np.linalg.norm(lin_vel, axis=1)
        lin_vel_orient = (lin_vel.T / lin_vel_mag).T
        if self.hierarchy.im_info.no_z:
            lin_vel_orient = np.where(np.isinf(lin_vel_orient), [np.nan, np.nan], lin_vel_orient)
        else:
            lin_vel_orient = np.where(np.isinf(lin_vel_orient), [np.nan, np.nan, np.nan], lin_vel_orient)

        return lin_vel, lin_vel_mag, lin_vel_orient

    def _get_angular_velocity_2d(self, ra, rb):
        # calculate angles of ra and rb relative to x-axis
        theta_a = np.arctan2(ra[:, 1], ra[:, 0])
        theta_b = np.arctan2(rb[:, 1], rb[:, 0])

        # calculate the change in angle
        delta_theta = theta_b - theta_a

        # normalize the change in angle to be between -pi and pi
        delta_theta = (delta_theta + np.pi) % (2 * np.pi) - np.pi

        # get the angular velocity
        ang_vel = delta_theta / self.hierarchy.im_info.dim_sizes['T']
        ang_vel_mag = np.abs(ang_vel)
        ang_vel_orient = np.sign(ang_vel)

        return ang_vel, ang_vel_mag, ang_vel_orient

    def _get_angular_velocity_3d(self, ra, rb):
        cross_product = np.cross(ra, rb, axis=1)
        norm = np.linalg.norm(ra, axis=1) * np.linalg.norm(rb, axis=1)
        ang_disp = np.divide(cross_product.T, norm.T).T
        ang_disp[norm == 0] = [np.nan, np.nan, np.nan]

        ang_vel = ang_disp / self.hierarchy.im_info.dim_sizes['T']
        ang_vel_mag = np.linalg.norm(ang_vel, axis=1)
        ang_vel_orient = (ang_vel.T / ang_vel_mag).T
        ang_vel_orient = np.where(np.isinf(ang_vel_orient), [np.nan, np.nan, np.nan], ang_vel_orient)

        return ang_vel, ang_vel_mag, ang_vel_orient

    def _get_angular_velocity(self, ra, rb):
        if self.hierarchy.im_info.no_z:
            return self._get_angular_velocity_2d(ra, rb)

        return self._get_angular_velocity_3d(ra, rb)

    def _run_frame(self, t=None):
        frame_coords = np.argwhere(self.hierarchy.label_components[t] > 0)
        self.coords.append(frame_coords)

        frame_component_labels = self.hierarchy.label_components[t][tuple(frame_coords.T)]
        self.component_labels.append(frame_component_labels)

        frame_branch_labels = self.hierarchy.label_branches[t][tuple(frame_coords.T)]
        self.branch_labels.append(frame_branch_labels)

        frame_intensity_vals = self.hierarchy.im_raw[t][tuple(frame_coords.T)]
        self.intensity.append(frame_intensity_vals)

        frame_structure_vals = self.hierarchy.im_struct[t][tuple(frame_coords.T)]
        self.structure.append(frame_structure_vals)

        frame_t = np.ones(frame_coords.shape[0], dtype=int) * t
        self.time.append(frame_t)

        im_name = np.ones(frame_coords.shape[0], dtype=object) * self.hierarchy.im_info.basename_no_ext
        self.image_name.append(im_name)

        self._get_node_info(t, frame_coords)
        self._get_motility_stats(t, frame_coords)

    def run(self):
        if self.hierarchy.num_t is None:
            self.hierarchy.num_t = 1
        for t in range(self.hierarchy.num_t):
            self._run_frame(t)


def aggregate_stats_for_class(child_class, t, list_of_idxs):
    # initialize a dictionary to hold lists of aggregated stats for each stat name
    aggregate_stats = {
        stat_name: {"mean": [], "std_dev": [], "25%": [], "50%": [], "75%": [], "min": [], "max": [], "range": [],
                    "sum": []} for
        stat_name in child_class.stats_to_aggregate if stat_name != 'reassigned_label'}

    for stat_name in child_class.stats_to_aggregate:
        if stat_name == 'reassigned_label':
            continue
        # access the relevant attribute for the current time frame
        stat_array = getattr(child_class, stat_name)[t]

        for idxs in list_of_idxs:
            stat_values = np.array(stat_array)[idxs]
            if stat_values.size == 0:
                # if there are no values for the current node, append a list of nans to each list in the aggregate_stats dictionary
                for key in aggregate_stats[stat_name].keys():
                    aggregate_stats[stat_name][key].append(np.nan)
                continue

            # calculate various statistics for the subset
            mean = np.nanmean(stat_values, axis=0)
            std_dev = np.nanstd(stat_values, axis=0)
            quartiles = np.nanquantile(stat_values, [0.25, 0.5, 0.75], axis=0)
            min_val = np.nanmin(stat_values, axis=0)
            max_val = np.nanmax(stat_values, axis=0)
            range_val = max_val - min_val
            sum_val = np.nansum(stat_values, axis=0)

            # append the calculated statistics to their respective lists in the aggregate_stats dictionary
            aggregate_stats[stat_name]["mean"].append(mean)
            aggregate_stats[stat_name]["std_dev"].append(std_dev)
            aggregate_stats[stat_name]["25%"].append(quartiles[0])
            aggregate_stats[stat_name]["50%"].append(quartiles[1])  # median
            aggregate_stats[stat_name]["75%"].append(quartiles[2])
            aggregate_stats[stat_name]["min"].append(min_val)
            aggregate_stats[stat_name]["max"].append(max_val)
            aggregate_stats[stat_name]["range"].append(range_val)
            aggregate_stats[stat_name]["sum"].append(sum_val)

    return aggregate_stats


class Nodes:
    def __init__(self, hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.nodes = []

        # add voxel aggregate metrics
        self.aggregate_voxel_metrics = []
        # add node metrics
        self.thickness = []
        self.divergence = []
        self.convergence = []
        self.vergere = []
        self.lin_magnitude_variability = []
        self.ang_magnitude_variability = []
        self.lin_direction_uniformity = []
        self.ang_direction_uniformity = []

        self.stats_to_aggregate = [
            "divergence", "convergence", "vergere", "lin_magnitude_variability", "ang_magnitude_variability",
            "lin_direction_uniformity", "ang_direction_uniformity", "thickness",
        ]

        self.features_to_save = self.stats_to_aggregate

        self.voxel_idxs = self.hierarchy.voxels.node_voxel_idxs
        self.branch_label = []
        self.component_label = []
        self.image_name = []

        self.node_z_lims = self.hierarchy.voxels.node_dim0_lims
        self.node_y_lims = self.hierarchy.voxels.node_dim1_lims
        self.node_x_lims = self.hierarchy.voxels.node_dim2_lims

    def _get_aggregate_voxel_stats(self, t):
        frame_agg = aggregate_stats_for_class(self.hierarchy.voxels, t, self.hierarchy.voxels.node_voxel_idxs[t])
        self.aggregate_voxel_metrics.append(frame_agg)

    def _get_node_stats(self, t):
        radius = distance_check(self.hierarchy.im_border_mask[t], self.nodes[t], self.hierarchy.spacing)
        self.thickness.append(radius * 2)

        divergence = []
        convergence = []
        vergere = []
        lin_mag_variability = []
        ang_mag_variability = []
        lin_dir_uniformity = []
        ang_dir_uniformity = []
        for i, node in enumerate(self.nodes[t]):
            vox_idxs = self.voxel_idxs[t][i]
            dist_vox_node = self.hierarchy.voxels.coords[t][vox_idxs] - self.nodes[t][i]
            dist_vox_node_mag = np.linalg.norm(dist_vox_node, axis=1, keepdims=True)
            dir_vox_node = dist_vox_node / dist_vox_node_mag

            dot_prod_01 = -np.nanmean(np.sum(-self.hierarchy.voxels.vec01[t][vox_idxs] * dir_vox_node, axis=1))
            convergence.append(dot_prod_01)

            dot_prod_12 = np.nanmean(np.sum(self.hierarchy.voxels.vec12[t][vox_idxs] * dir_vox_node, axis=1))
            divergence.append(dot_prod_12)

            vergere.append(dot_prod_01 + dot_prod_12)
            # high vergere is a funnel point (converges then diverges)
            # low vergere is a dispersal point (diverges then converges)

            lin_vel_mag = self.hierarchy.voxels.lin_vel_mag[t][vox_idxs]
            lin_mag_variability.append(np.nanstd(lin_vel_mag))
            ang_vel_mag = self.hierarchy.voxels.ang_vel_mag[t][vox_idxs]
            ang_mag_variability.append(np.nanstd(ang_vel_mag))

            lin_vel = self.hierarchy.voxels.lin_vel[t][vox_idxs]
            lin_unit_vec = lin_vel / lin_vel_mag[:, np.newaxis]
            lin_similarity_matrix = np.dot(lin_unit_vec, lin_unit_vec.T)
            np.fill_diagonal(lin_similarity_matrix, np.nan)
            lin_dir_uniformity.append(np.nanmean(lin_similarity_matrix))

            ang_vel = self.hierarchy.voxels.ang_vel[t][vox_idxs]
            ang_unit_vec = ang_vel / ang_vel_mag[:, np.newaxis]
            ang_similarity_matrix = np.dot(ang_unit_vec, ang_unit_vec.T)
            np.fill_diagonal(ang_similarity_matrix, np.nan)
            ang_dir_uniformity.append(np.nanmean(ang_similarity_matrix))
        self.divergence.append(divergence)
        self.convergence.append(convergence)
        self.vergere.append(vergere)
        self.lin_magnitude_variability.append(lin_mag_variability)
        self.ang_magnitude_variability.append(ang_mag_variability)
        self.lin_direction_uniformity.append(lin_dir_uniformity)
        self.ang_direction_uniformity.append(ang_dir_uniformity)

    def _run_frame(self, t):
        frame_skel_coords = np.argwhere(self.hierarchy.im_pixel_class[t] > 0)
        self.nodes.append(frame_skel_coords)

        frame_t = np.ones(frame_skel_coords.shape[0], dtype=int) * t
        self.time.append(frame_t)

        frame_component_label = self.hierarchy.label_components[t][tuple(frame_skel_coords.T)]
        self.component_label.append(frame_component_label)

        frame_branch_label = self.hierarchy.label_branches[t][tuple(frame_skel_coords.T)]
        self.branch_label.append(frame_branch_label)

        im_name = np.ones(frame_skel_coords.shape[0], dtype=object) * self.hierarchy.im_info.basename_no_ext
        self.image_name.append(im_name)

        self._get_aggregate_voxel_stats(t)
        self._get_node_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            self._run_frame(t)


def distance_check(border_mask, check_coords, spacing):
    border_coords = np.argwhere(border_mask) * spacing
    border_tree = spatial.cKDTree(border_coords)
    dist, _ = border_tree.query(check_coords * spacing, k=1)
    return dist


class Branches:
    def __init__(self, hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.branch_label = []

        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []
        # add branch metrics
        self.length = []
        self.thickness = []
        self.aspect_ratio = []
        self.tortuosity = []
        self.area = []
        self.axis_length_maj = []
        self.axis_length_min = []
        self.extent = []
        self.solidity = []
        self.reassigned_label = []

        self.branch_idxs = []
        self.component_label = []
        self.image_name = []

        self.stats_to_aggregate = [
            "length", "thickness", "aspect_ratio", "tortuosity", "area", "axis_length_maj", "axis_length_min",
            "extent", "solidity", "reassigned_label"
        ]

    def _get_aggregate_stats(self, t):
        voxel_labels = self.hierarchy.voxels.branch_labels[t]
        grouped_vox_idxs = [np.argwhere(voxel_labels == label).flatten()
                            for label in np.unique(voxel_labels) if label != 0]
        vox_agg = aggregate_stats_for_class(self.hierarchy.voxels, t, grouped_vox_idxs)
        self.aggregate_voxel_metrics.append(vox_agg)

        node_labels = self.hierarchy.nodes.branch_label[t]
        grouped_node_idxs = [np.argwhere(node_labels == label).flatten()
                             for label in np.unique(node_labels) if label != 0]
        node_agg = aggregate_stats_for_class(self.hierarchy.nodes, t, grouped_node_idxs)
        self.aggregate_node_metrics.append(node_agg)

    def _get_branch_stats(self, t):
        branch_idx_array_1 = np.array(self.branch_idxs[t])
        branch_idx_array_2 = np.array(self.branch_idxs[t])[:, None, :]
        dist = np.linalg.norm(branch_idx_array_1 - branch_idx_array_2, axis=-1)
        dist[dist >= 2] = 0  # remove any distances greater than adjacent pixel
        neighbors = np.sum(dist > 0, axis=1)
        tips = np.where(neighbors == 1)[0]
        lone_tips = np.where(neighbors == 0)[0]
        dist = np.triu(dist)

        neighbor_idxs = np.argwhere(dist > 0)

        # all coords idxs should be within 1 pixel of each other
        neighbor_coords_0 = self.branch_idxs[t][neighbor_idxs[:, 0]]
        neighbor_coords_1 = self.branch_idxs[t][neighbor_idxs[:, 1]]
        assert np.all(np.abs(neighbor_coords_0 - neighbor_coords_1) <= 1)

        # labels should be the exact same
        neighbor_labels_0 = self.hierarchy.im_skel[t][tuple(neighbor_coords_0.T)]
        neighbor_labels_1 = self.hierarchy.im_skel[t][tuple(neighbor_coords_1.T)]
        assert np.all(neighbor_labels_0 == neighbor_labels_1)

        scaled_coords_0 = neighbor_coords_0 * self.hierarchy.spacing
        scaled_coords_1 = neighbor_coords_1 * self.hierarchy.spacing
        distances = np.linalg.norm(scaled_coords_0 - scaled_coords_1, axis=1)
        unique_labels = np.unique(self.hierarchy.im_skel[t][self.hierarchy.im_skel[t] > 0])

        label_lengths = np.zeros(len(unique_labels))
        for i, label in enumerate(unique_labels):
            label_lengths[i] = np.sum(distances[neighbor_labels_0 == label])

        lone_tip_coords = self.branch_idxs[t][lone_tips]
        tip_coords = self.branch_idxs[t][tips]

        lone_tip_labels = self.hierarchy.im_skel[t][tuple(lone_tip_coords.T)]
        tip_labels = self.hierarchy.im_skel[t][tuple(tip_coords.T)]

        # find the distance between the two tips with the same label in tip_labels
        tip_distances = np.zeros(len(tip_labels))
        for i, label in enumerate(tip_labels):
            matched_idxs = tip_coords[tip_labels == label] * self.hierarchy.spacing
            tip_distances[i] = np.linalg.norm(matched_idxs[0] - matched_idxs[1])

        tortuosity = np.zeros(len(unique_labels))
        for i, label in enumerate(unique_labels):
            tip_dist = tip_distances[tip_labels == label]
            if len(tip_dist):
                tortuosity[i] = label_lengths[i] / tip_dist[0]
            else:
                tortuosity[i] = 1

        self.tortuosity.append(tortuosity)

        radii = distance_check(self.hierarchy.im_border_mask[t], self.branch_idxs[t], self.hierarchy.spacing)
        lone_tip_radii = radii[lone_tips]
        tip_radii = radii[tips]

        for label, radius in zip(lone_tip_labels, lone_tip_radii):
            label_lengths[unique_labels == label] += radius * 2
        for label, radius in zip(tip_labels, tip_radii):
            label_lengths[unique_labels == label] += radius

        # mean radii for each branch:
        median_thickenss = []
        thicknesses = radii * 2
        for label, thickness in zip(unique_labels, thicknesses):
            median_thickenss.append(np.median(thickness))

        # if thickness at an index is larger than the length, set it to the length, and length to thickness
        for i, thickness in enumerate(median_thickenss):
            if thickness > label_lengths[i]:
                median_thickenss[i] = label_lengths[i]
                label_lengths[i] = thickness

        aspect_ratios = label_lengths / median_thickenss

        self.aspect_ratio.append(aspect_ratios)
        self.thickness.append(median_thickenss)
        self.length.append(label_lengths)

        regions = regionprops(self.hierarchy.label_branches[t], spacing=self.hierarchy.spacing)
        areas = []
        axis_length_maj = []
        axis_length_min = []
        extent = []
        solidity = []
        reassigned_label = []

        for region in regions:
            if not self.hierarchy.im_info.no_t:
                region_reassigned_labels = self.hierarchy.im_branch_reassigned[t][tuple(region.coords.T)]
                # find which label is most common in the region via bin-counting
                reassigned_label.append(np.argmax(np.bincount(region_reassigned_labels)))
            else:
                reassigned_label.append(np.nan)
            areas.append(region.area)
            # due to bug in skimage (at the time of writing: https://github.com/scikit-image/scikit-image/issues/6630)
            try:
                maj_axis = region.major_axis_length
                min_axis = region.minor_axis_length
            except ValueError:
                maj_axis = np.nan
                min_axis = np.nan
            axis_length_maj.append(maj_axis)
            axis_length_min.append(min_axis)
            extent.append(region.extent)
            solidity.append(region.solidity)
        self.area.append(areas)
        self.axis_length_maj.append(axis_length_maj)
        self.axis_length_min.append(axis_length_min)
        self.extent.append(extent)
        self.solidity.append(solidity)
        self.reassigned_label.append(reassigned_label)

    def _run_frame(self, t):
        frame_branch_idxs = np.argwhere(self.hierarchy.im_skel[t] > 0)
        self.branch_idxs.append(frame_branch_idxs)

        frame_skel_branch_labels = self.hierarchy.im_skel[t][tuple(frame_branch_idxs.T)]

        smallest_label = int(np.min(self.hierarchy.im_skel[t][self.hierarchy.im_skel[t] > 0]))
        largest_label = int(np.max(self.hierarchy.im_skel[t]))
        frame_branch_labels = np.arange(smallest_label, largest_label + 1)
        num_branches = len(frame_branch_labels)

        frame_t = np.ones(num_branches, dtype=int) * t
        self.time.append(frame_t)

        # get the first voxel idx for each branch
        if self.hierarchy.im_info.no_z:
            frame_branch_coords = np.zeros((num_branches, 2), dtype=int)
        else:
            frame_branch_coords = np.zeros((num_branches, 3), dtype=int)
        for i in frame_branch_labels:
            branch_voxels = frame_branch_idxs[frame_skel_branch_labels == i]
            frame_branch_coords[i - 1] = branch_voxels[0]
        frame_component_label = self.hierarchy.label_components[t][tuple(frame_branch_coords.T)]
        self.component_label.append(frame_component_label)

        frame_branch_label = self.hierarchy.im_skel[t][tuple(frame_branch_coords.T)]
        self.branch_label.append(frame_branch_label)

        im_name = np.ones(num_branches, dtype=object) * self.hierarchy.im_info.basename_no_ext
        self.image_name.append(im_name)

        self._get_aggregate_stats(t)
        self._get_branch_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            self._run_frame(t)


class Components:
    def __init__(self, hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.component_label = []
        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []
        self.aggregate_branch_metrics = []
        # add component metrics
        self.area = []
        self.axis_length_maj = []
        self.axis_length_min = []
        self.extent = []
        self.solidity = []
        self.reassigned_label = []

        self.image_name = []

        self.stats_to_aggregate = [
            "area", "axis_length_maj", "axis_length_min", "extent", "solidity", "reassigned_label",
        ]

    def _get_aggregate_stats(self, t):
        voxel_labels = self.hierarchy.voxels.component_labels[t]
        grouped_vox_idxs = [np.argwhere(voxel_labels == label).flatten() for label in np.unique(voxel_labels) if
                            label != 0]
        vox_agg = aggregate_stats_for_class(self.hierarchy.voxels, t, grouped_vox_idxs)
        self.aggregate_voxel_metrics.append(vox_agg)

        node_labels = self.hierarchy.nodes.component_label[t]
        grouped_node_idxs = [np.argwhere(node_labels == label).flatten() for label in np.unique(voxel_labels) if
                             label != 0]
        node_agg = aggregate_stats_for_class(self.hierarchy.nodes, t, grouped_node_idxs)
        self.aggregate_node_metrics.append(node_agg)

        branch_labels = self.hierarchy.branches.component_label[t]
        grouped_branch_idxs = [np.argwhere(branch_labels == label).flatten() for label in np.unique(voxel_labels) if
                               label != 0]
        branch_agg = aggregate_stats_for_class(self.hierarchy.branches, t, grouped_branch_idxs)
        self.aggregate_branch_metrics.append(branch_agg)

    def _get_component_stats(self, t):
        regions = regionprops(self.hierarchy.label_components[t], spacing=self.hierarchy.spacing)
        areas = []
        axis_length_maj = []
        axis_length_min = []
        extent = []
        solidity = []
        reassigned_label = []

        for region in regions:
            if not self.hierarchy.im_info.no_t:
                region_reassigned_labels = self.hierarchy.im_obj_reassigned[t][tuple(region.coords.T)]
                reassigned_label.append(np.argmax(np.bincount(region_reassigned_labels)))
            else:
                reassigned_label.append(np.nan)
            areas.append(region.area)
            try:
                maj_axis = region.major_axis_length
                min_axis = region.minor_axis_length
            except ValueError:
                maj_axis = np.nan
                min_axis = np.nan
            axis_length_maj.append(maj_axis)
            axis_length_min.append(min_axis)
            extent.append(region.extent)
            solidity.append(region.solidity)
        self.area.append(areas)
        self.axis_length_maj.append(axis_length_maj)
        self.axis_length_min.append(axis_length_min)
        self.extent.append(extent)
        self.solidity.append(solidity)
        self.reassigned_label.append(reassigned_label)

    def _run_frame(self, t):
        smallest_label = int(np.min(self.hierarchy.label_components[t][self.hierarchy.label_components[t] > 0]))
        largest_label = int(np.max(self.hierarchy.label_components[t]))
        frame_component_labels = np.arange(smallest_label, largest_label + 1)
        self.component_label.append(frame_component_labels)
        num_components = len(frame_component_labels)

        frame_t = np.ones(num_components, dtype=int) * t
        self.time.append(frame_t)

        im_name = np.ones(num_components, dtype=object) * self.hierarchy.im_info.basename_no_ext
        self.image_name.append(im_name)

        self._get_aggregate_stats(t)
        self._get_component_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            self._run_frame(t)


class Image:
    def __init__(self, hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.image_name = []
        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []
        self.aggregate_branch_metrics = []
        self.aggregate_component_metrics = []
        self.stats_to_aggregate = []

    def _get_aggregate_stats(self, t):
        voxel_agg = aggregate_stats_for_class(self.hierarchy.voxels, t,
                                              [np.arange(len(self.hierarchy.voxels.coords[t]))])
        self.aggregate_voxel_metrics.append(voxel_agg)

        node_agg = aggregate_stats_for_class(self.hierarchy.nodes, t, [np.arange(len(self.hierarchy.nodes.nodes[t]))])
        self.aggregate_node_metrics.append(node_agg)

        branch_agg = aggregate_stats_for_class(self.hierarchy.branches, t,
                                               [self.hierarchy.branches.branch_label[t].flatten() - 1])
        self.aggregate_branch_metrics.append(branch_agg)

        component_agg = aggregate_stats_for_class(self.hierarchy.components, t,
                                                  [np.arange(len(self.hierarchy.components.component_label[t]))])
        self.aggregate_component_metrics.append(component_agg)

    def _run_frame(self, t):
        self.time.append(t)
        self.image_name.append(self.hierarchy.im_info.basename_no_ext)

        self._get_aggregate_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            self._run_frame(t)


if __name__ == "__main__":
    im_path = "/Users/austin/test_files/nelly/ND Stimulation Parallel 12.nd2"
    im_info = ImInfo(im_path)
    num_t = 3

    hierarchy = Hierarchy(im_info, num_t)
    hierarchy.run()



import numpy as np

from nellie.im_info.im_info import ImInfo
from nellie.tracking.flow_interpolation import interpolate_all_forward, interpolate_all_backward
from nellie.utils.general import get_reshaped_image


class LabelTracks:
    def __init__(self, im_info: ImInfo, num_t: int = None, label_im_path: str = None):
        self.im_info = im_info
        self.num_t = num_t
        if label_im_path is None:
            label_im_path = self.im_info.pipeline_paths['im_instance_label']
        self.label_im_path = label_im_path

        if num_t is None:
            self.num_t = im_info.shape[im_info.axes.index('T')]

    def initialize(self):
        self.label_memmap = self.im_info.get_im_memmap(self.label_im_path)
        self.label_memmap = get_reshaped_image(self.label_memmap, im_info=self.im_info)
        self.im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(self.im_memmap, im_info=self.im_info)

    def run(self, label_num=None, start_frame=0, end_frame=None, min_track_num=0, skip_coords=1, max_distance_um=0.5):
        if end_frame is None:
            end_frame = self.num_t
        num_frames = self.label_memmap.shape[0] - 1
        if start_frame > num_frames:
            return [], {}
        if label_num is None:
            coords = np.argwhere(self.label_memmap[start_frame] > 0).astype(float)
        else:
            coords = np.argwhere(self.label_memmap[start_frame] == label_num).astype(float)
        if coords.shape[0] == 0:
            return [], {}
        coords = np.array(coords[::skip_coords])
        coords_copy = coords.copy()
        tracks = []
        track_properties = {}
        if start_frame < end_frame:
            tracks, track_properties = interpolate_all_forward(coords, start_frame, end_frame, self.im_info,
                                                               min_track_num, max_distance_um=max_distance_um)
        new_end_frame = 0  # max(0, end_frame - start_frame)
        if start_frame > 0:
            tracks_bw, track_properties_bw = interpolate_all_backward(coords_copy, start_frame, new_end_frame,
                                                                      self.im_info, min_track_num,
                                                                      max_distance_um=max_distance_um)
            tracks_bw = tracks_bw[::-1]
            for property in track_properties_bw.keys():
                track_properties_bw[property] = track_properties_bw[property][::-1]
            sort_idx = np.argsort([track[0] for track in tracks_bw])
            tracks_bw = [tracks_bw[i] for i in sort_idx]
            tracks = tracks_bw + tracks
            for property in track_properties_bw.keys():
                track_properties_bw[property] = [track_properties_bw[property][i] for i in sort_idx]
            if not track_properties:
                track_properties = track_properties_bw
            else:
                for property in track_properties_bw.keys():
                    track_properties[property] = track_properties_bw[property] + track_properties[property]
        return tracks, track_properties


if __name__ == "__main__":
    im_path = r"D:\test_files\nellie_longer_smorgasbord\deskewed-peroxisome.ome.tif"
    im_info = ImInfo(im_path, ch=0)
    num_t = 20
    label_tracks = LabelTracks(im_info, num_t=num_t)
    label_tracks.initialize()
    # tracks, track_properties = label_tracks.run(label_num=None, skip_coords=1)

    all_tracks = []
    all_props = {}
    max_track_num = 0
    for frame in range(num_t):
        tracks, track_properties = label_tracks.run(label_num=None, start_frame=frame, end_frame=None,
                                                    min_track_num=max_track_num,
                                                    skip_coords=100)
        all_tracks += tracks
        for property in track_properties.keys():
            if property not in all_props.keys():
                all_props[property] = []
            all_props[property] += track_properties[property]
        if len(tracks) == 0:
            break
        max_track_num = max([track[0] for track in tracks]) + 1

    # pickle tracks and track properties
    import pickle
    import datetime
    dt = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(f'{dt}-mt_tracks.pkl', 'wb') as f:
        pickle.dump(all_tracks, f)
    with open(f'{dt}-mt_props.pkl', 'wb') as f:
        pickle.dump(all_props, f)

    # import napari
    # viewer = napari.Viewer()
    #
    # raw_im = im_info.get_im_memmap(im_info.im_path)[:num_t]
    # viewer.add_image(raw_im, name='raw_im')
    # viewer.add_tracks(all_tracks, properties=all_props, name='tracks')
    # napari.run()
